from time import time
from random import randint
from nitro.async import AsyncOp
from nitro.upnp import UPnpActions
from cspace.util.statemachine import StateMachine

class UPnpMapper( object ) :
    DEFAULT = 0
    INITIALIZING = 1
    READY = 2
    CLEANINGUP = 3
    CLOSED = 4

    def __init__( self, reactor ) :
        self.reactor = reactor
        self.sm = StateMachine( self.DEFAULT )
        self.device = None
        self.externalIP = None
        self.cleaningupCount = 0
        self.sm.appendCallback( self._onCleanup, dest=self.CLEANINGUP, single=True )
        self._initialize()

    def shutdown( self, callback=None ) :
        if self.sm.current() in (self.CLEANINGUP,self.CLOSED) :
            return None
        if self.sm.current() == self.READY :
            self.sm.change( self.CLEANINGUP )
        else :
            self.sm.change( self.CLOSED )
        if self.sm.current() == self.CLOSED :
            return None
        def onClosed() : op.notify()
        def doCancel() : self.sm.removeCallback( callbackId )
        callbackId = self.sm.appendCallback( onClosed,
                dest=self.CLOSED, single=True )
        op = AsyncOp( callback, doCancel )
        return op

    def _onCleanup( self ) :
        if self.cleaningupCount == 0 :
            self.sm.change( self.CLOSED )

    def _initialize( self ) :
        class Dummy : pass
        obj = Dummy()
        def onError() :
            self.sm.removeCallback( callbackId )
            self.sm.change( self.CLOSED )
        def onDiscover( device ) :
            if device is None :
                onError()
                return
            self.device = device
            obj.op = UPnpActions.getExternalIP( device, self.reactor,
                    onExternalIP )
        def onExternalIP( externalIP ) :
            if externalIP is None :
                onError()
                return
            self.externalIP = externalIP
            self.sm.removeCallback( callbackId )
            self.sm.change( self.READY )
        self.sm.change( self.INITIALIZING )
        obj.op = UPnpActions.findDevice( self.reactor, onDiscover )
        callbackId = self.sm.insertCallback( lambda : obj.op.cancel(),
                src=self.INITIALIZING, single=True )

    def addMapping( self, localIP, localPort, callback=None ) :
        class Dummy : pass
        obj = Dummy()
        if self.sm.current() not in (self.INITIALIZING,self.READY) :
            timerOp = self.reactor.callLater( 0, lambda : obj.op.notify(None) )
            obj.op = AsyncOp( callback, timerOp.cancel )
            return obj.op
        def doReady() :
            obj.attempt = 0
            doAttempt()
        def doAttempt() :
            if obj.attempt == 3 :
                obj.op.notify( None )
                return
            obj.attempt += 1
            obj.externalPort = randint( 10000, 20000 )
            desc = 'CSpace_t%d' % int(time())
            obj.addOp = UPnpActions.addMapping( self.device, obj.externalPort,
                    'TCP', localPort, localIP, desc, self.reactor, onAdd )
            obj.callbackId = self.sm.insertCallback( onAbort, src=self.READY,
                    single=True )
            obj.op.setCanceler( onCancel )
        def onCancel() :
            obj.addOp.cancel()
            self.sm.removeCallback( obj.callbackId )
        def onAbort() :
            obj.addOp.cancel()
            obj.op.notify( None )
        def onAdd( result ) :
            self.sm.removeCallback( obj.callbackId )
            if not result :
                doAttempt()
                return
            mapping = (self.externalIP,obj.externalPort)
            obj.op.notify( mapping )
            self.sm.insertCallback( onCleanup, dest=self.CLEANINGUP, single=True )
        def onCleanup() :
            self.cleaningupCount += 1
            UPnpActions.delMapping( self.device, obj.externalPort, 'TCP',
                    self.reactor, onDelMapping )
        def onDelMapping( result ) :
            self.cleaningupCount -= 1
            self._onCleanup()
        if self.sm.current() == self.INITIALIZING :
            def checkReady() :
                if self.sm.current() == self.READY :
                    obj.op.setCanceler( None )
                    doReady()
                    return
                obj.op.notify( None )
            obj.callbackId = self.sm.insertCallback( checkReady,
                    src=self.INITIALIZING, single=True )
            obj.op = AsyncOp( callback,
                    lambda : self.sm.removeCallback(obj.callbackId) )
            return obj.op
        obj.op = AsyncOp( callback, None )
        doReady()
        return obj.op



#!/usr/bin/python
#backup_fs.py
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#


import os
import sys
import stat
import platform
import hashlib
import base64
import cStringIO
import time
import locale

try:
    import lib.dhnio as dhnio
except:
    dirpath = os.path.dirname(os.path.abspath(sys.argv[0]))
    sys.path.insert(0, os.path.abspath('datahaven'))
    sys.path.insert(0, os.path.abspath(os.path.join(dirpath, '..')))
    sys.path.insert(0, os.path.abspath(os.path.join(dirpath, '..', '..')))
    try:
        import lib.dhnio as dhnio
    except:
        sys.exit()
        
import lib.misc as misc
import lib.settings as settings        
import lib.packetid as packetid

#------------------------------------------------------------------------------ 

INFO_KEY = 'i'
UNKNOWN = -1
FILE = 0
DIR = 1
PARENT = 2
TYPES = {UNKNOWN:   'UNKNOWN',
         FILE:      'FILE',
         DIR:       'DIR',
         PARENT:    'PARENT',} 

#------------------------------------------------------------------------------ 

_FileSystemIndexByName = {}
_FileSystemIndexByID = {}
_ItemsCount = 0
_FilesCount = 0
_SizeFiles = 0
_SizeBackups = 0

def fs():
    global _FileSystemIndexByName
    return _FileSystemIndexByName

def fsID():
    global _FileSystemIndexByID
    return _FileSystemIndexByID

def counter():
    global _ItemsCount
    return _ItemsCount

def numberfiles():
    global _FilesCount
    return _FilesCount

def sizefiles():
    global _SizeFiles
    return _SizeFiles

def sizebackups():
    global _SizeBackups
    return _SizeBackups

#------------------------------------------------------------------------------ 

def init():
    dhnio.Dprint(4, 'backup_fs.init')
    # fs()[settings.BackupIndexFileName()] = -1
    # fsID()[-1] = FSItemInfo(settings.BackupIndexFileName(), '-1', FILE)
    # backup_fs.SetFile(settings.BackupIndexFileName(), settings.BackupIndexFileName())

def shutdown():
    dhnio.Dprint(4, 'backup_fs.shutdown')

#-------------------------------------------------------------i----------------- 

class FSItemInfo():
    def __init__(self, name='', path='', typ=UNKNOWN):
        if isinstance(name, unicode):
            self.unicodename = name
        else:
            self.unicodename = unicode(name)
        self.path = path
        self.type = typ
        self.size = -1
        self.versions = {} # set()

    def name(self): 
        return self.unicodename

    def exist(self):
        return self.size != -1
        
    def read_stats(self, path):
        if not pathExist(path):
            self.size = -1
            return
        try:
            s = os.stat(path)
        except:
            try:
                s = os.stat(path.decode('utf-8'))
            except:
                dhnio.DprintException()
                self.size = -1
                return
        self.size = long(s.st_size)
        
    def read_versions(self, path):
        path = portablePath(path)
        if not pathExist(path):
            return 0
        if not os.access(path, os.R_OK):
            return 0
        totalSize = 0
        for version in os.listdir(path):
            if self.get_version_info(version)[0] >= 0:
                continue
            versionSize = 0
            maxBlock = -1
            if not packetid.IsCanonicalVersion(version):
                continue
            versionpath = os.path.join(path, version)
            if not pathExist(versionpath):
                continue
            if not os.access(versionpath, os.R_OK):
                return 0
            for filename in os.listdir(versionpath):
                filepath = os.path.join(versionpath, filename)
                if not packetid.IsPacketNameCorrect(filename):
                    dhnio.Dprint(4, 'backup_fs.read_versions WARNING incorrect file name found: %s' % filepath)
                    continue
                try:
                    blockNum, supplierNum, dataORparity = filename.split('-')
                    blockNum, supplierNum = int(blockNum), int(supplierNum)
                except:
                    dhnio.Dprint(4, 'backup_fs.read_versions WARNING incorrect file name found: %s' % filepath)
                    continue
                try:
                    sz = long(os.path.getsize(filepath))
                except:
                    dhnio.DprintException()
                # add some bytes because on remote machines all packets are dhnpackets
                # so they have a header and the files size will be bigger than on local machine
                versionSize += sz + 1024   
                maxBlock = max(maxBlock, blockNum)
            self.set_version_info(version, maxBlock, versionSize)
            totalSize += versionSize
        return totalSize
        
    def add_version(self, version):
        # self.versions.add(version)
        self.versions[version] = [-1, -1] # max block num, size in bytes
        
    def set_version_info(self, version, maxblocknum, sizebytes):
        self.versions[version] = [maxblocknum, sizebytes]
        
    def get_version_info(self, version):
        return self.versions.get(version, [-1, -1])
        
    def delete_version(self, version):
        # self.versions.remove(version)
        self.versions.pop(version, None)
        
    def has_version(self, version):
        # return version in self.versions
        return self.versions.has_key(version)
    
    def any_version(self):
        return len(self.versions) > 0
    
    def list_versions(self, sorted=False, reverse=False):
        # return list(self.versions)
        if sorted:
            return misc.sorted_versions(self.versions.keys(), reverse)
        return self.versions.keys()
    
    def get_versions(self):
        return self.versions
    
    def pack_versions(self):
        out = []
        for version in self.list_versions(True):
            info = self.versions[version]
            out.append(version+':'+str(info[0])+':'+str(info[1]))
        return ' '.join(out)
    
    def unpack_versions(self, input):
        for word in input.split(' '):
            if not word.strip():
                continue
            try:
                version, maxblock, sz = word.split(':')
                maxblock, sz = int(maxblock), int(sz)
            except:
                version, maxblock, sz = word, -1, -1
            self.set_version_info(version, maxblock, sz)
        
    def serialize(self, encoding='utf-8'):
        e = self.unicodename.encode(encoding)
        return '%s %d %d %s\n%s\n' % (self.path, 
                                         self.type, 
                                         self.size,
                                         self.pack_versions(),
                                         e,)
        
    def unserialize(self, src, decoding='utf-8'):
        try:
            details, name = src.split('\n')[:2]
        except:
            raise
        if details == '' or name == '':
            raise Exception('Incorrect item format:\n%s' % src) 
        try:
            self.unicodename = name.decode(decoding)
            details = details.split(' ')
            self.path, self.type, self.size = details[:3]
            self.type, self.size = int(self.type), int(self.size)
            self.unpack_versions(' '.join(details[3:])) 
        except:
            raise

    def __repr__(self):
        return '<%s %s %d>' % (TYPES[self.type], misc.unicode_to_str_safe(self.name()), self.size)

#------------------------------------------------------------------------------ 

def portablePath(path):
    """ fix path to fit for our use:
            + do convert to absolute path
            + for Windows: + change all separators to Linux format: \\->/ \=>/
                           + convert disk letter to lower case """
    p = os.path.abspath(path)
    if not isinstance(p, unicode):
        # p = p.encode('utf-8')
        p = unicode(p)
    if dhnio.Windows():
        p = p.replace('\\', '/') # .replace('\\\\', '/')
        if len(p) >= 2 and p[1] == ':':
            p = p[0].lower() + p[1:]
    if p.endswith('/') and len(p) > 1:
        p = p.rstrip('/')
    return p # unicode(p) #.encode('utf-8')

# our own portable version of os.path.exist
def pathExist(localpath):
    if os.path.exists(localpath):
        return True
    p = portablePath(localpath)
    if os.path.exists(p):
        return True
#    if os.path.exists(p.decode('utf-8')):
#        return True
    return False

# assume localpath is exist and is absolute or raise Exception
def pathIsDir(localpath):
    # print 'pathIsDir', type(localpath), str(localpath)
    # try to use the original path
    if os.path.isdir(localpath):
        return True
    if os.path.isfile(localpath):
        return False
    # don't know... let's try portable path
    p = portablePath(localpath)
    if os.path.isdir(p):
        return True
    if os.path.isfile(p):
        return False
    # may be path is not exist at all?
    if not os.path.exists(localpath):
        return False
    if not os.path.exists(p):
        return False
    # ok, on Linux we have devices, mounts, links ...
    if dhnio.Linux():
        try:
            st = os.path.stat(localpath)
            return stat.S_ISDIR(st.st_mode)
        except:
            return False 
    # now we are in really big trouble
    raise Exception('Path not exist: %s' % p)
    return False

def pathIsDriveLetter(path):
    p = path.rstrip('/').rstrip('\\')
    if len(p) != 2:
        return False
    if p[1] != ':':
        return False
    if not p[0].isalpha():
        return False
    return True

#------------------------------------------------------------------------------ 

def MakeID(iter, startID=-1):
    id = 0
    if startID >= 0:
        id = startID
    while id in map(lambda k: '' if k == 0 else (iter[k] if isinstance(iter[k], int) else iter[k][0]), iter.keys()):
        id += 1
    return id

#------------------------------------------------------------------------------ 

def AddFile(path, read_stats=False, iter=None, iterID=None):
    # here path must be portable - only '/' allowed
    # assume path is a file, not dir
    if not os.path.isfile(path):
        raise Exception('File not exist')
    parts = portablePath(path).split('/')
    if not iter:
        iter = fs()
    if not iterID:
        iterID = fsID()
    resultID = ''
    # build all tree, skip the last part
    for i in range(len(parts)-1):
        name = parts[i]
        if not name:
            continue
        p = '/'.join(parts[:i+1])
        if not pathIsDir(p):
            raise Exception('Directory not exist: %s' % str(p))
        if not iter.has_key(name):
            # made a new ID for this folder, ID starts from 0. new folders will get the last ID +1 
            # or it may find a free place in the middle, if some folders or files were removed before
            # this way we try to protect the files and directories names. we store index in the encrypted files
            id = MakeID(iter)
            # build a unique backup id for that file including all indexed ids
            resultID += '/' + str(id)
            # make new sub folder 
            ii = FSItemInfo(name, resultID.lstrip('/'), PARENT)
            if read_stats:
                ii.read_stats(p)
            # we use 0 key as decimal value, all files and folders are strings - no conflicts possible 0 != '0'
            iter[ii.name()] = {0: id}
            # also save index from opposite side
            iterID[id] = {INFO_KEY: ii}
        else:
            # get an existing ID from the index
            id = iter[name][0]
            # go down into the existing forest
            resultID += '/' + str(id)
        # move down to the next level
        iter = iter[name]
        iterID = iterID[id]
    # the last part of the path is a filename
    filename = parts[-1] 
    # make an ID for the filename
    id = MakeID(iter)
    resultID += '/' + str(id)
    ii = FSItemInfo(filename, resultID, FILE)
    if read_stats:
        ii.read_stats(path)
    iter[ii.name()] = id
    iterID[id] = ii
    # finally make a complete backup id - this a relative path to the backed up file
    return resultID.lstrip('/'), iter, iterID

def AddDir(path, read_stats=False, iter=None, iterID=None):
    parts = portablePath(path).split('/')
    if not iter:
        iter = fs()
    if not iterID:
        iterID = fsID()
    resultID = ''
    for i in range(len(parts)):
        name = parts[i]
        if not name:
            continue
        p = '/'.join(parts[:i+1])
        if not pathIsDir(p):
            raise Exception('Directory not exist: %s' % str(p))
        if not iter.has_key(name):
            id = MakeID(iter)
            resultID += '/' + str(id)
            ii = FSItemInfo(name, resultID.lstrip('/'), PARENT)
            if read_stats:
                ii.read_stats(p) 
            iter[ii.name()] = {0: id}
            iterID[id] = {INFO_KEY: ii}
        else:
            id = iter[name][0]
            resultID += '/' + str(id)
        iter = iter[name]
        iterID = iterID[id]
        if i == len(parts) - 1:
            iterID[INFO_KEY].type = DIR
    return resultID.lstrip('/'), iter, iterID

def AddLocalPath(localpath, read_stats=False):
    def recursive_read_dir(path, path_id, iter, iterID):
        c = 0
        lastID = -1
        path = portablePath(path)
        if not os.access(path, os.R_OK):
            return c
        for localname in os.listdir(path):
            p = os.path.join(path, localname)  # .encode("utf-8")
            name = unicode(localname) 
            if pathIsDir(p):
                if not iter.has_key(name):
                    id = MakeID(iter, lastID)
                    ii = FSItemInfo(name, path_id+'/'+str(id), DIR)
                    iter[ii.name()] = {0: id}
                    if read_stats:
                        ii.read_stats(p)
                    iterID[id] = {INFO_KEY: ii}
                    lastID = id
                else:
                    id = iter[name][0]
                c += recursive_read_dir(p, path_id+'/'+str(id), iter[name], iterID[id])
            else:
                id = MakeID(iter, lastID)
                ii = FSItemInfo(name, path_id+'/'+str(id), FILE)
                if read_stats:
                    ii.read_stats(p)
                iter[ii.name()] = id
                iterID[id] = ii
                c += 1
                lastID = id
        return c    
    localpath = portablePath(localpath) 
    if pathIsDir(localpath):
        path_id, iter, iterID = AddDir(localpath, read_stats)
        num = recursive_read_dir(localpath, path_id, iter, iterID)
        return path_id, iter, iterID, num
    else:
        path_id, iter, iterID = AddFile(localpath, read_stats)
        return path_id, iter, iterID, 1
    return None, None, None, 0

#------------------------------------------------------------------------------ 

def SetFile(item, iter=None, iterID=None):
    if iter is None:
        iter = fs()
    if iterID is None:
        iterID = fsID()
    parts = item.path.split('/')
    for j in range(len(parts)):
        id = int(parts[j])
        if j == len(parts) - 1:
            if not iter.has_key(item.name()):
                iter[item.name()] = id
                iterID[id] = item
            return True
        found = False
        for name in iter.keys():
            if name == 0:
                continue
            if isinstance(iter[name], dict):
                if iter[name][0] == id:
                    iter = iter[name]
                    iterID = iterID[id]
                    found = True
                    break
                continue
        if not found:
            return False
    return False

def SetDir(item, iter=None, iterID=None):
    if iter is None:
        iter = fs()
    if iterID is None:
        iterID = fsID()
    parts = item.path.split('/')
    itemname = item.name()
    for j in range(len(parts)):
        id = int(parts[j])
        if j == len(parts)-1:
            if not iter.has_key(itemname):
                iter[itemname] = {}
            iter[itemname][0] = int(id)
            if not iterID.has_key(id):
                iterID[id] = {}
            iterID[id][INFO_KEY] = item
            return True
        found = False
        for name in iter.keys():
            if name == 0:
                continue
            if isinstance(iter[name], int):
                continue
            if isinstance(iter[name], dict):
                if iter[name][0] == id:
                    iter = iter[name]
                    iterID = iterID[id]
                    found = True
                    break
                continue
            raise Exception('Wrong data type in the index')
        if not found:
            return False
    return False

#------------------------------------------------------------------------------ 

def WalkByPath(path, iter=None):
    if iter is None:
        iter = fs()
    path_id = ''
    parts = portablePath(path).split('/')
    for j in range(len(parts)):
        name = parts[j]
        if not iter.has_key(name):
            return None
        if isinstance(iter[name], dict):
            if not iter[name].has_key(0):
                raise Exception('Error, file or directory ID missed in the index')
            path_id += '/' + str(iter[name][0])
        elif isinstance(iter[name], int):
            if j != len(parts) - 1:
                return None
            path_id += '/' + str(iter[name])
        else:
            raise Exception('Wrong data type in the index')
        if j == len(parts) - 1:
            return iter[name], path_id.lstrip('/')
        iter = iter[name]
    return None   

def WalkByID(pathID, iterID=None):
    if iterID is None:
        iterID = fsID()
    if pathID is None:
        return None
    if pathID.strip() == '':
        return iterID, ''
    path = ''
    parts = pathID.strip('/').split('/')
    for j in range(len(parts)):
        try:
            id = int(parts[j])
        except:
            return None
        if not iterID.has_key(id):
            return None
        if isinstance(iterID[id], dict):
            if not iterID[id].has_key(INFO_KEY):
                raise Exception('Error, directory info missed in the index')
            name = iterID[id][INFO_KEY].name()
            if pathIsDriveLetter(name):
                path += name
            else:
                path += '/' + name 
        elif isinstance(iterID[id], FSItemInfo):
            if j != len(parts) - 1:
                return None
            path += '/' + iterID[id].name()
        else:
            raise Exception('Wrong data type in the index')
        if j == len(parts) - 1:
            return iterID[id], path
        iterID = iterID[id]
    return None

#------------------------------------------------------------------------------ 

def DeleteByID(pathID, iter=None, iterID=None):
    if iter is None:
        iter = fs()
    if iterID is None:
        iterID = fsID()
    path = ''
    parts = pathID.strip('/').split('/')
    for j in range(len(parts)):
        id = int(parts[j])
        if not iterID.has_key(id):
            return None
        if isinstance(iterID[id], dict):
            if not iterID[id].has_key(INFO_KEY):
                raise Exception('Error, directory info missed in the index')
            name = iterID[id][INFO_KEY].name()
            if pathIsDriveLetter(name):
                path += name
            else:
                path += '/' + name 
        elif isinstance(iterID[id], FSItemInfo):
            name = iterID[id].name()
            path += '/' + name
            if j != len(parts) - 1:
                return None
        else:
            raise Exception('Wrong data type in the index')
        if not iter.has_key(name):
            raise Exception('Can not found target name in the index')
        if j == len(parts) - 1:
            iterID.pop(id)
            iter.pop(name)
            return path
        iterID = iterID[id]
        iter = iter[name]
    return None
    
def DeleteByPath(path, iter=None, iterID=None):
    if iter is None:
        iter = fs()
    if iterID is None:
        iterID = fsID()
    path_id = ''
    parts = portablePath(path).split('/')
    for j in range(len(parts)):
        name = parts[j]  # .encode('utf-8') # parts[j]
        if not iter.has_key(name):
            return None
        if isinstance(iter[name], dict):
            if not iter[name].has_key(0):
                raise Exception('Error, directory ID missed in the index')
            id = iter[name][0]
            path_id += '/' + str(id)
        elif isinstance(iter[name], int):
            id = iter[name]
            path_id += '/' + str(id)
            if j != len(parts) - 1:
                return None
        else:
            raise Exception('Wrong data type in the index')
        if not iterID.has_key(id):
            raise Exception('Can not found target ID in the index')
        if j == len(parts) - 1:
            iter.pop(name)
            iterID.pop(id)
            return path_id.lstrip('/')
        iter = iter[name]
        iterID = iterID[id]
    return None

def DeleteBackupID(backupID):
    pathID, versionName = packetid.SplitBackupID(backupID)
    if pathID is None:
        return False
    info = GetByID(pathID)
    if info is None:
        return False
    if not info.has_version(versionName):
        dhnio.Dprint(4, 'backup_fs.DeleteBackupID WARNING %s do not have version %s' % (pathID, versionName))
        return False 
    info.delete_version(versionName)
    return True
    
#------------------------------------------------------------------------------ 

def ToID(localPath, iter=None):
    iter_and_id = WalkByPath(localPath, iter)
    if iter_and_id is None:
        return None
    return iter_and_id[1]

def ToPath(pathID, iterID=None):
    """ get a full relative path from ID 
    /0/0/12/1 -> /home/veselin/Documents/somefile.txt
    return None if that path not exist
    """
    iter_and_path = WalkByID(pathID, iterID=None)
    if iter_and_path is None:
        return None
    return iter_and_path[1]

#------------------------------------------------------------------------------ 

def GetByID(pathID):
    iter_and_path = WalkByID(pathID)
    if iter_and_path is None:
        return None
    if isinstance(iter_and_path[0], dict):
        return iter_and_path[0][INFO_KEY]
    return iter_and_path[0]

def GetByPath(path):
    path_id = ToID(path)
    if path_id is None:
        return None
    return GetByID(path_id)

#------------------------------------------------------------------------------ 

def IsDir(path):
    iter_and_id = WalkByPath(path)
    if iter_and_id is None:
        return False
    iter, pathid = iter_and_id
    if not isinstance(iter, dict):
        return False
    if not iter.has_key(0):
        raise Exception('Error, directory ID missed in the index')
    return True    

def IsFile(path):
    iter_and_id = WalkByPath(path)
    if iter_and_id is None:
        return False
    iter, pathid = iter_and_id
    if not isinstance(iter, int):
        return False
    return True    

def IsDirID(pathID):
    iter_and_path = WalkByID(pathID)
    if iter_and_path is None:
        return False
    iterID, path = iter_and_path
    if isinstance(iterID, FSItemInfo):
        return False
    if not iterID.has_key(INFO_KEY):
        raise Exception('Error, directory info missed in the index')
    return True

def IsFileID(pathID):
    iter_and_path = WalkByID(pathID)
    if iter_and_path is None:
        return False
    iterID, path = iter_and_path
    if not isinstance(iterID, FSItemInfo):
        return False
    return True
    

def Exists(path):
    iter_and_id = WalkByPath(path)
    if iter_and_id is None:
        return False
    return True    

def ExistsID(pathID):
    iter_and_path = WalkByID(pathID)
    if iter_and_path is None:
        return False
    return True   

def ExistsBackupID(backupID):
    pathID, version = packetid.SplitBackupID(backupID)
    if not pathID:
        return False
    iter_and_path = WalkByID(pathID)
    if iter_and_path is None:
        return False
    iter, path = iter_and_path
    return iter.has_version(version)

#------------------------------------------------------------------------------ 

def HasChilds(path):
    iter_and_id = WalkByPath(path)
    if iter_and_id is None:
        return False
    iter, pathid = iter_and_id
    if not isinstance(iter, dict):
        return False
    if not iter.has_key(0):
        raise Exception('Error, directory ID missed in the index')
    return len(iter) > 1

def HasChildsID(pathID):
    iter_and_path = WalkByID(pathID)
    if iter_and_path is None:
        return False
    iterID, path = iter_and_path
    if not isinstance(iterID, dict):
        return False
    if not iterID.has_key(INFO_KEY):
        raise Exception('Error, directory info missed in the index')
    return len(iterID) > 1
    
#------------------------------------------------------------------------------ 

def TraverseByID(callback, iterID=None): 
    '''     call callback(path_id, path, info) for every item in the index    '''
    def recursive_traverse(i, path_id, path, cb):
        name = None
        if path != '':
            path += '/'
        if isinstance(i, FSItemInfo):
            cb(path_id, path, i)
            return
        if i.has_key(INFO_KEY):
            name = i[INFO_KEY].name()
            path += name
            cb(path_id, path, i[INFO_KEY])
        for id in i.keys():
            if id == INFO_KEY:
                continue
            if isinstance(i[id], dict):
                recursive_traverse(i[id], path_id+'/'+str(id) if path_id else str(id), path, cb)
            elif isinstance(i[id], FSItemInfo):
                cb(path_id+'/'+str(id) if path_id else str(id), path+'/'+i[id].name() if path else i[id].name(), i[id])
            else:
                raise Exception('Error, wrong item type in the index')
    if iterID is None:
        iterID = fsID()
    startpth = '' if dhnio.Windows() else '/'
    recursive_traverse(iterID, '', startpth, callback)
    
def TraverseByIDSorted(callback, iterID=None): 
    def recursive_traverse(i, path_id, path, cb):
        name = None
        if path != '':
            path += '/'
        if i.has_key(INFO_KEY):
            name = i[INFO_KEY].name()
            path += name
            cb(path_id, path, i[INFO_KEY])
        dirs = []
        files = []
        for id in i.keys():
            if id == INFO_KEY:
                continue
            if isinstance(i[id], dict):
                dirs.append((id, path+'/'+i[id][INFO_KEY].name() if path else i[id][INFO_KEY].name()))
            elif isinstance(i[id], FSItemInfo):
                files.append((id, path+'/'+i[id].name() if path else i[id].name()))
            else:
                raise Exception('Error, wrong item type in the index')   
        dirs.sort(key=lambda e: e[1])
        files.sort(key=lambda e: e[1])
        for id, pth in dirs:
            recursive_traverse(i[id], path_id+'/'+str(id) if path_id else str(id), path, cb)
        for id, pth in files:
            cb(path_id+'/'+str(id) if path_id else str(id), pth, i[id])
        del dirs
        del files
    if iterID is None:
        iterID = fsID()
    startpth = '' if dhnio.Windows() else '/'
    recursive_traverse(iterID, '', startpth, callback)
    
def IterateIDs(iterID=None):
    # iterator is (pathID, localPath, item)
    if iterID is None:
        iterID = fsID()
    def recursive_iterate(i, path_id, path):
        name = None
        if path != '':
            path += '/'
        if i.has_key(INFO_KEY):
            name = i[INFO_KEY].name()
            path += name
            yield path_id, path, i[INFO_KEY]
        for id in i.keys():
            if id == INFO_KEY:
                continue
            if isinstance(i[id], dict):
                for t in recursive_iterate(i[id], path_id+'/'+str(id) if path_id else str(id), path): 
                    yield t
            elif isinstance(i[id], FSItemInfo):
                yield path_id+'/'+str(id) if path_id else str(id), path+'/'+i[id].name(), i[id]
            else:
                raise Exception('Error, wrong item type in the index')        
    startpth = '' if dhnio.Windows() else '/'
    return recursive_iterate(iterID, '', startpth)    

#------------------------------------------------------------------------------ 

def ListByPath(path, iter=None):
    path = portablePath(path)
    iter_and_id = WalkByPath(path, iter)
    if iter_and_id is None:
        return None
    result = []
    iter, path_id = iter_and_id
    if isinstance(iter, int):
        return [(path_id, path),]
    if not isinstance(iter, dict):
        raise Exception('Wrong data type in the index')
    if not iter.has_key(0):
        raise Exception('Error, directory ID missed in the index')
    for key in iter.keys():
        if key == 0:
            continue
        if isinstance(iter[key], dict):
            if not iter[key].has_key(0):
                raise Exception('Error, directory ID missed in the index')
            id = iter[key][0]
        elif isinstance(iter[key], int):
            id = iter[key]
        else:
            raise Exception('Wrong data type in the index')
        result.append((path_id+'/'+str(id), path+'/'+key))
    return result    
    
def ListByID(pathID, iterID=None):  
    iter_and_path = WalkByID(pathID, iterID)
    if iter_and_path is None:
        return None
    result = []
    iterID, path = iter_and_path
    if isinstance(iterID, FSItemInfo):
        return [(pathID, path),]
    if not isinstance(iterID, dict):
        raise Exception('Wrong data type in the index')
    if not iterID.has_key(INFO_KEY) and pathID.strip() != '':
        raise Exception('Error, directory info missed in the index')
    for id in iterID.keys():
        if id == INFO_KEY:
            continue
        if isinstance(iterID[id], dict):
            if not iterID[id].has_key(INFO_KEY):
                raise Exception('Error, directory info missed in the index')
            name = iterID[id][INFO_KEY].name()
        elif isinstance(iterID[id], FSItemInfo):
            name = iterID[id].name()
        else:
            raise Exception('Wrong data type in the index')
        result.append((pathID+'/'+str(id), path+'/'+name))
    return result    

def ListAllBackupIDs(sorted=False, reverse=False, iterID=None):
    lst = []
    def visitor(path_id, path, info):
        for version in info.list_versions(sorted, reverse):
            lst.append(str(path_id + '/' + version))
    TraverseByID(visitor)
    return lst

def ListAllBackupIDsFull(sorted=False, reverse=False, iterID=None):
    lst = []
    def visitor(path_id, path, info):
        for version in info.list_versions(sorted, reverse):
            lst.append((str(path_id + '/' + version), info.get_version_info(version), path))
    TraverseByID(visitor)
    return lst

def ListSelectedFolders(selected_dirs_ids, sorted=False, reverse=False):
    lst = []
    def visitor(path_id, path, info):
        basepathid = path_id[:path_id.rfind('/')] if path_id.count('/') else ''
        if basepathid in selected_dirs_ids: 
            lst.append((info.type, path_id, path, info.size, info.list_versions(sorted, reverse)))
        return True
    TraverseByIDSorted(visitor)
    return lst

def ListExpandedFoldersAndBackups(expanded_dirs, selected_items):
    lst = []
    backups = []
    def visitor(path_id, path, info):
        basepathid = path_id[:path_id.rfind('/')] if path_id.count('/') else ''
        if basepathid in expanded_dirs: 
            lst.append((info.type, path_id, path, info.size, info.get_versions()))
        if path_id in selected_items:
            for version in info.list_versions():
                backups.append(path_id+'/'+version)
        return True
    TraverseByIDSorted(visitor)
    return lst, backups

#------------------------------------------------------------------------------ 

def MakeLocalDir(basedir, pathID):
    if not pathIsDir(basedir):
        raise Exception('Directory not exist: %s' % basedir)
    path = os.path.join(basedir, pathID)
    if os.path.exists(path):
        if not pathIsDir(path):
            raise Exception('Can not create directory %s' % path)
    else:
        os.makedirs(path)
    return path

def DeleteLocalDir(basedir, pathID):
    if not pathIsDir(basedir):
        raise Exception('Directory not exist: %s' % basedir)
    path = os.path.join(basedir, pathID)
    if not os.path.exists(path):
        return
    if not pathIsDir(path):
        raise Exception('Error, %s is not a directory' % path)
    dhnio.rmdir_recursive(path)
    
def DeleteLocalBackup(basedir, backupID):
    count_and_size = [0, 0,]
    if not pathIsDir(basedir):
        raise Exception('Directory not exist: %s' % basedir)
    backupDir = os.path.join(basedir, backupID)
    if not pathExist(backupDir):
        return count_and_size[0], count_and_size[1]
    if not pathIsDir(backupDir):
        raise Exception('Error, %s is not a directory' % backupDir)
    def visitor(fullpath):
        if os.path.isfile(fullpath):
            count_and_size[0] += 1
            count_and_size[1] += os.path.getsize(fullpath) 
        return True
    dhnio.rmdir_recursive(backupDir, pre_callback=visitor)
    return count_and_size[0], count_and_size[1]

#------------------------------------------------------------------------------ 

def Scan():
    iterID = fsID()
    sum = [0, 0,]
    def visitor(path_id, path, info):
        info.read_stats(path)
        if info.exist():
            sum[0] += info.size
        versions_path = portablePath(os.path.join(settings.getLocalBackupsDir(), path_id))
        sum[1] += info.read_versions(versions_path)
    TraverseByID(visitor, iterID)
    return sum[0], sum[1]

def ScanID(pathID):
    iter_and_path = WalkByID(pathID)
    if not iter_and_path:
        return
    iter, path = iter_and_path
    if isinstance(iter, dict):
        if not iter.has_key(INFO_KEY):
            return
        iter = iter[INFO_KEY]
    iter.read_stats(path)
    iter.read_versions(portablePath(os.path.join(settings.getLocalBackupsDir(), pathID)))
    
def Calculate():
    global _SizeFiles
    global _SizeBackups
    global _ItemsCount
    global _FilesCount
    _ItemsCount = 0
    _FilesCount = 0
    _SizeFiles = 0
    _SizeBackups = 0
    def recursive_calculate(i):
        global _SizeFiles
        global _SizeBackups
        global _ItemsCount
        global _FilesCount
        folder_size = 0
        for id in i.keys():
            if id == INFO_KEY:
                continue
            if isinstance(i[id], FSItemInfo):
                if i[id].exist():
                    folder_size += i[id].size
                if i[id].type == FILE:
                    _FilesCount += 1
                    if i[id].exist():
                        _SizeFiles += i[id].size
                for version in i[id].list_versions():
                    versionSize = i[id].get_version_info(version)[1]
                    if versionSize > 0:
                        _SizeBackups += versionSize 
                _ItemsCount += 1
            elif isinstance(i[id], dict):
                sub_folder_size = recursive_calculate(i[id])
                if sub_folder_size != -1:
                    folder_size += sub_folder_size
            else:
                raise Exception('Error, wrong item type in the index')
        if i.has_key(INFO_KEY):
            i[INFO_KEY].size = folder_size
            if i[INFO_KEY].type == FILE:
                _FilesCount += 1
                if i[INFO_KEY].exist():
                    _SizeFiles += i[INFO_KEY].size
            for version in i[INFO_KEY].list_versions():
                versionSize = i[INFO_KEY].get_version_info(version)[1]
                if versionSize > 0:
                    _SizeBackups += versionSize 
            _ItemsCount += 1
        return folder_size
    return recursive_calculate(fsID())

def Calculate2(iterID=None):
    if iterID is None:
        iterID = fsID()
    def recursive_calculate(i):
        folder_size = 0
        for id in i.keys():
            if id == INFO_KEY:
                continue
            if isinstance(i[id], FSItemInfo):
                if i[id].exist():
                    folder_size += i[id].size
            elif isinstance(i[id], dict):
                sub_folder_size = recursive_calculate(i[id])
                if sub_folder_size != -1:
                    folder_size += sub_folder_size
            else:
                raise Exception('Error, wrong item type in the index')
        if i.has_key(INFO_KEY):
            i[INFO_KEY].size = folder_size
        return folder_size
    return recursive_calculate(iterID)
    
#------------------------------------------------------------------------------ 

def Clear():
    fs().clear()
    fsID().clear()

def Serialize(iterID=None):
    result = cStringIO.StringIO()
    cnt = [0]
    def cb(path_id, path, info):
        result.write(info.serialize())
        cnt[0] += 1
    TraverseByID(cb, iterID)
    src = result.getvalue()
    result.close()
    dhnio.Dprint(6, 'backup_fs.Serialize done with %d indexed files' % cnt[0])
    return src

def Unserialize(input, iter=None, iterID=None):
    count = 0
    while True:
        src = input.readline() + input.readline() # 2 times because we take 2 lines for every item
        if src.strip() == '':
            break
        item = FSItemInfo()
        item.unserialize(src)
        if item.type == FILE:
            if not SetFile(item, iter, iterID):
                raise Exception('Can not put item into the tree: %s' % str(item))
        elif item.type == DIR or item.type == PARENT:
            if not SetDir(item, iter, iterID):
                raise Exception('Can not put item into the tree: %s' % str(item))
        else:
            raise Exception('Incorrect entry type')
        count += 1
    dhnio.Dprint(6, 'backup_fs.Unserialize done with %d indexed files' % count)
    return count

#------------------------------------------------------------------------------ 

def main():
    import pprint
#    dhnio.init()
#    for path in sys.argv[1:]:
#        print path, AddLocalPath(path, True)
#
#    for pathID, localPath, itemInfo in IterateIDs():
#        print pathID, localPath, itemInfo
#        
#    print IsDirID('0/0/0')

    # pprint.pprint(fs())
    # pprint.pprint(fsID())
    # s = Serialize()
    # print s
#    open('index', 'w').write(s)
#    for id, path in ListByID(''):
#        pprint.pprint(ListByID(id))
#    fs().clear()
#    fsID().clear()
#    inp = cStringIO.StringIO(s)
#    Unserialize(inp)
#    inp.close()
#    pprint.pprint(fs())
#    pprint.pprint(fsID())
    
    # print GetPath('/0/0/0/1/31/12')
    # print GetID('C:/.Veselin/diplom/Social/Editor/type 5.vsd.aut')
    # print WalkByPath('C:/.Veselin/diplom/Social/Editor/type 5.vsd.aut')
    # print WalkByID('/0/0/0/1/31/12')
    # print IsDir('C:')
    # DeleteByID('/0')
    # print ListByID('')
    # print 'NEW C:/.Veselin/diplom/', AddLocalPath('C:/.Veselin/diplom/')
    # print ListByID('')
    # pprint.pprint(ListByPath('c:\.Veselin/diplom/'))
    # print DeleteByID('0/0/0/0')
    # print DeleteByPath('C:/.Veselin/diplom/Social')
    # pprint.pprint(fs())
    # pprint.pprint(ListByPath('C:/.Veselin/diplom/Social/Editor'))
    
    # iter = fs()['c:']['work']['modnamama']['_katalog']
    # print map(lambda k: iter[k][0] if isinstance(iter, dict) else '', iter.keys())
    # pprint.pprint( ListLocalFolder(sys.argv[1]) )
    # print ListLocalFolder(sys.argv[1])

if __name__ == '__main__':
    main()

    



from types import IntType, LongType, StringType, ListType, TupleType, DictType

def encodeInt( x, out ) :
    assert type(x) in (IntType,LongType)
    out.extend( ('i', str(x), 'e') )

def encodeString( x, out ) :
    assert type(x) is StringType
    out.extend( (str(len(x)), ':', x) )

def encodeList( x, out ) :
    assert type(x) in (ListType,TupleType)
    out.append( 'l' )
    for i in x : encoderTable[type(i)]( i, out )
    out.append( 'e' )

def encodeDict( x, out ) :
    assert type(x) is DictType
    out.append( 'd' )
    items = x.items()
    items.sort()
    for a,b in items :
        encodeString( a, out )
        encoderTable[type(b)]( b, out )
    out.append( 'e' )

encoderTable = {}
encoderTable[IntType] = encodeInt
encoderTable[LongType] = encodeInt
encoderTable[StringType] = encodeString
encoderTable[ListType] = encodeList
encoderTable[TupleType] = encodeList
encoderTable[DictType] = encodeDict

def encode( x ) :
    out = []
    encoderTable[type(x)]( x, out )
    return ''.join( out )

class DecodeError( Exception ) : pass

def decodeInt( x, i ) :
    assert x[i] == 'i'
    e = x.index( 'e', i )
    i += 1
    ret = int(x[i:e])
    if x[i] == '-' :
        if x[i+1] == '0' : raise ValueError
    elif x[i] == '0' :
        if e != i+1 : raise ValueError
    return (ret, e+1)

def decodeString( x, i ) :
    e = x.index( ':', i )
    count = int( x[i:e] )
    if count < 0 : raise ValueError
    if x[i] == '0' and e != i+1 : raise ValueError
    e += 1
    ret = x[e:e+count]
    return (ret, e+count)

def decodeList( x, i ) :
    assert x[i] == 'l'
    ret, i = [], i+1
    next = x[i]
    while next != 'e' :
        (v, i) = decoderTable[next]( x, i )
        ret.append( v )
        next = x[i]
    return (ret, i+1)

def decodeDict( x, i ) :
    assert x[i] == 'd'
    ret, i = {}, i+1
    prev = None
    while x[i] != 'e' :
        (a, i) = decodeString( x, i )
        if a <= prev : raise ValueError
        prev = a
        (b, i) = decoderTable[x[i]]( x, i )
        ret[a] = b
    return (ret, i+1)

decoderTable = {}
decoderTable['i'] = decodeInt
for i in range(10) :
    decoderTable[chr(ord('0')+i)] = decodeString
decoderTable['l'] = decodeList
decoderTable['d'] = decodeDict

def decode( x ) :
    try :
        (ret,i) = decoderTable[x[0]]( x, 0 )
    except (KeyError, IndexError, ValueError) :
        raise DecodeError, 'invalid bencoded data'
    if i != len(x) :
        raise DecodeError, 'invalid tail data'
    return ret



#!/usr/bin/python
#restore_monitor.py
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#
#
#manage currently restoring backups

import os
import sys
import time

from twisted.internet import reactor, threads


import lib.dhnio as dhnio
import lib.tmpfile as tmpfile

import restore
import backup_tar
import backup_matrix

#------------------------------------------------------------------------------ 

_WorkingBackupIDs = {}
_WorkingRestoreProgress = {}
OnRestorePacketFunc = None
OnRestoreDoneFunc = None
OnRestoreBlockFunc = None

#------------------------------------------------------------------------------ 

def init():
    dhnio.Dprint(4, 'restore_monitor.init')


def block_restored_callback(backupID, block):
    global OnRestoreBlockFunc
    if OnRestoreBlockFunc is not None:
        OnRestoreBlockFunc(backupID, block)


def packet_in_callback(backupID, packet):
    # dhnio.Dprint(8, 'restore_monitor.packet_in_callback ' + backupID)
    global _WorkingRestoreProgress
    global OnRestorePacketFunc
    
    SupplierNumber = packet.SupplierNumber()
    
    #want to count the data we restoring
    if SupplierNumber not in _WorkingRestoreProgress[backupID].keys():
        _WorkingRestoreProgress[backupID][SupplierNumber] = 0
    _WorkingRestoreProgress[backupID][SupplierNumber] += len(packet.Payload)
    
    backup_matrix.LocalFileReport(packet.PacketID)
    
    if OnRestorePacketFunc is not None:
        OnRestorePacketFunc(backupID, SupplierNumber, packet)


def extract_done(retcode, backupID, tarfilename, callback):
    dhnio.Dprint(8, 'restore_monitor.extract_done %s result: %s' % (backupID, str(retcode)))
    global OnRestoreDoneFunc
    
    _WorkingBackupIDs.pop(backupID, None)
    _WorkingRestoreProgress.pop(backupID, None)

    tmpfile.throw_out(tarfilename, 'file extracted')
    
    if OnRestoreDoneFunc is not None:
        OnRestoreDoneFunc(backupID, 'restore done')

    if callback:
        callback(backupID, 'restore done')


def restore_done(x, tarfilename, outputlocation, callback):
    dhnio.Dprint(8, 'restore_monitor.restore_done ' + str(x))
    global _WorkingBackupIDs
    global _WorkingRestoreProgress
    global OnRestoreDoneFunc
    
    backupID, result = x.split(' ')
    
    if result == 'done':
        p = backup_tar.extracttar(tarfilename, outputlocation)
        if p:
            d = threads.deferToThread(p.wait)
            d.addBoth(extract_done, backupID, tarfilename, callback)
            return
        result = 'extract failed'

    _WorkingBackupIDs.pop(backupID, None)
    _WorkingRestoreProgress.pop(backupID, None)

    tmpfile.throw_out(tarfilename, 'restore '+result)

    if OnRestoreDoneFunc is not None:
        OnRestoreDoneFunc(backupID, result)
    
    if callback:
        callback(backupID, result)


def restore_failed(x, tarfilename, callback):
    dhnio.Dprint(8, 'restore_monitor.restore_failed ' + str(x))
    global _WorkingBackupIDs
    global _WorkingRestoreProgress
    global OnRestoreDoneFunc

    if isinstance(x, Exception) or isinstance(x, str):
        backupID, result = str(x).split(' ')
        
        _WorkingBackupIDs.pop(backupID, None)
        _WorkingRestoreProgress.pop(backupID, None)
        
        tmpfile.throw_out(tarfilename, 'restore '+result)
    
        if OnRestoreDoneFunc is not None:
            OnRestoreDoneFunc(backupID, result)
        
        if callback:
            callback(backupID, result) 


def Start(backupID, outputLocation, callback=None):
    dhnio.Dprint(8, 'restore_monitor.Start %s to %s' % (backupID, outputLocation))
    global _WorkingBackupIDs
    global _WorkingRestoreProgress
    if backupID in _WorkingBackupIDs.keys():
        return None
    outfd, outfilename = tmpfile.make('restore', '.tar.gz', backupID.replace('/','_')+'_')
    r = restore.restore(backupID, outfd)
    r.MyDeferred.addCallback(restore_done, outfilename, outputLocation, callback)
    r.MyDeferred.addErrback(restore_failed, outfilename, callback)
    r.SetBlockRestoredCallback(block_restored_callback)
    r.SetPacketInCallback(packet_in_callback)
    _WorkingBackupIDs[backupID] = r
    _WorkingRestoreProgress[backupID] = {}
    return r


def Abort(backupID):
    dhnio.Dprint(8, 'restore_monitor.Abort %s' % backupID)
    global _WorkingBackupIDs
    global _WorkingRestoreProgress
    if not backupID in _WorkingBackupIDs.keys():
        return False
    r = _WorkingBackupIDs[backupID]
    r.Abort()
    return True


def GetWorkingIDs():
    global _WorkingBackupIDs
    return _WorkingBackupIDs.keys()


def IsWorking(backupID):
    global _WorkingBackupIDs
    return backupID in _WorkingBackupIDs.keys()


def GetProgress(backupID):
    global _WorkingRestoreProgress
    return _WorkingRestoreProgress.get(backupID, {})


def GetWorkingRestoreObject(backupID):
    global _WorkingBackupIDs
    return _WorkingBackupIDs.get(backupID, None)





from hexcode import hexByteEncode, hexByteDecode, HexDecodeError

class WordDecodeError( Exception ) : pass

def wordEncode( s ) :
    out = []
    for c in s :
        if (not c.isalnum()) and (c not in '_-.') :
            out.append( '%' )
            out.append( hexByteEncode(c) )
        else :
            out.append( c )
    return ''.join( out )

def wordDecode( s ) :
    try :
        out = []
        i = 0
        while i < len(s) :
            if s[i] == '%' :
                out.append( hexByteDecode(s,i+1) )
                i += 3
            else :
                out.append( s[i] )
                i += 1
        return ''.join( out )
    except HexDecodeError :
        raise WordDecodeError



from distutils.core import setup

setup(  name='datahaven',
        version='rev8249',
        author='Vincent Cate, DataHaven.NET LTD',
        author_email='datahaven.net@gmail.com',
        maintainer='Veselin Penev',
        maintainer_email='penev.veselin@gmail.com',
        url='http://datahaven.net',
        description='Distributed encrypted backup utility',
        long_description='''DataHaven.NET is a peer-to-peer online backup utility.
        
This is a distributed network for backup data storage. 
Each participant of the network provides a portion of his hard drive for other users. 
In exchange, he is able to store his data on other users's machines.

The redundancy in backups makes it so if someone loses your data, 
you can reconstruct what was lost and give it to someone else to hold. 
And all of this happens without you having to do a thing.

All data is encrypted before it leaves your computer with a key your computer generates. 
No one else can read your data. Recover data is only one way - 
download the necessary pieces from computers of other users and 
decrypt them with your private key.
        
The system is designed in such a way that will 
keep track of each user which stores your information, and each block of your
data and maintain a state in which you can always get your data back.

You can imagine it like this: you exchange two gigabytes in a safe place
(on your hard drive) to one gigabyte, but in the distributed storage 
and so kept very reliable. If you are going to store data 
that you are not constantly in use,
but you need that they would be most protected - this is what you need!
 
This is very similar to the well-known torrents... but on the contrary! :-)))

''',
        download_url='http://datahaven.net',
        #download_url='http://pypi.python.org/packages/source/d/datahaven/datahaven-rev8249.tar.gz',
        license='Copyright DataHaven.NET LTD. of Anguilla, 2006-2012,\nAll rights reserved.\nUse of this software constitutes acceptance of the Terms of Use: http://datahaven.net/terms_of_use.html ',
        keywords='''p2p, peer to peer, backup, restore, distributed, cloud, online, storage, data, recover, python, twisted''',
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Environment :: Console',
            'Environment :: Web Environment',
            'Framework :: Twisted',
            'Intended Audience :: Customer Service',
            'Intended Audience :: Developers',
            'Intended Audience :: End Users/Desktop',
            'Intended Audience :: Information Technology',
            'Intended Audience :: System Administrators',
            'License :: Other/Proprietary License',
            'Natural Language :: English',
            'Operating System :: Microsoft :: Windows',
            'Operating System :: Microsoft :: Windows :: Windows 7',
            'Operating System :: Microsoft :: Windows :: Windows Vista',
            'Operating System :: Microsoft :: Windows :: Windows XP',
            'Operating System :: POSIX :: Linux',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2.5',
            'Programming Language :: Python :: 2.6',
            'Programming Language :: Python :: 2.7',
            'Topic :: Security',
            'Topic :: Security :: Cryptography',
            'Topic :: System :: Archiving :: Backup',
            'Topic :: System :: Distributed Computing',
            'Topic :: Utilities',
            ],
            
        packages=[
            'datahaven',
            'datahaven.p2p', 
            'datahaven.forms',
            'datahaven.lib',
            'datahaven.lib.shtoom',            
            'datahaven.cspace',
            'datahaven.cspace.dht',
            'datahaven.cspace.main',
            'datahaven.cspace.network',
            'datahaven.cspace.util',
            'datahaven.nitro',
            'datahaven.ncrypt0',
            ],
)







#!/usr/bin/python
#tmpfile.py
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#


import os
import tempfile
import time


from twisted.internet import reactor, task


import dhnio


_TempDirPath = None
_FilesDict = {}
_CollectorTask = None
_SubDirs = {

    'outbox':   60*60*1,
    # hold onto outbox files 1 hour
    # so we can handle resends if contact is off-line

    # 'data-par': 0,
    # for data and parity files we have special rules
    # because both files types are stored in the same folder "data-par"
    # also backup_monitor seems to be responsible to remove these files
    # but we will remove old files to - at startup
    # hold onto data for rebuilding for a day, can rebuild parity as needed
    # hold onto parity for an hour

    'tcp-in':   60*10,
    # 10 minutes for incoming tcp files
    
    'ssh-in':   60*10,
    # 10 minutes for incoming ssh files

    'q2q-in':   60*10,
    # 10 minutes for incoming files

    'http-in':  60*30,
    # 30 minutes for incoming http files
    # http is very slow so let's give more time

    'email-in': 60*60,
    # 1 hour for incoming email files
    # email is so slow so we may be want to keep it more

    'cspace-in':60*10,
    # 10 minutes for incoming files
    
    'bat':      60*60,
    # we need bat files to restart dhn
    # we create it and start immediately
    # so we can remove it soon

    'propagat': 60*10,
    # propagate happens often enough,
    # 10 minutes should be enough

    'backup':   60*10,
    # 10 minutes for backup files
    
    'restore':  60*10,
    # 10 minutes for restoring files

    'raid':   60*10,
    # 10 minutes for backup files

    'other':    0,
    # other files. do not know when to remove
    # they can be even in another location
    # use register(name, filename)

    }



#------------------------------------------------------------------------------


def init(temp_dir_path=''):
    dhnio.Dprint(4, 'tmpfile.init')
    global _TempDirPath
    global _SubDirs
    global _FilesDict
    global _CollectorTask

    if _TempDirPath is None:
        if temp_dir_path != '':
            _TempDirPath = temp_dir_path
        else:
            os_temp_dir = tempfile.gettempdir()
            temp_dir = os.path.join(os_temp_dir, 'dhn')

            if not os.path.exists(temp_dir):
                try:
                    os.mkdir(temp_dir)
                except:
                    dhnio.Dprint(2, 'tmpfile.init ERROR can not create ' + temp_dir)
                    dhnio.DprintException()
                    temp_dir = os_temp_dir

            if not os.access(temp_dir, os.W_OK):
                dhnio.Dprint(2, 'tmpfile.init ERROR no write permissions to ' + temp_dir)
                temp_dir = os_temp_dir

            _TempDirPath = temp_dir
        dhnio.Dprint(6, 'tmpfile.init  _TempDirPath=' + _TempDirPath)

    for name in _SubDirs.keys():
        if not os.path.exists(subdir(name)):
            try:
                os.makedirs(subdir(name))
            except:
                dhnio.Dprint(2, 'tmpfile.init ERROR can not create ' + subdir(name))
                dhnio.DprintException()

    for name in _SubDirs.keys():
        if not _FilesDict.has_key(name):
            _FilesDict[name] = {}

    startup_clean()

    if _CollectorTask is None:
        _CollectorTask = task.LoopingCall(collect)
        _CollectorTask.start(60*5)


def shutdown():
    # we do not want to remove any files here
    # just stop the collector task
    dhnio.Dprint(4, 'tmpfile.shutdown')
    global _CollectorTask
    if _CollectorTask is not None:
        _CollectorTask.stop()
        del _CollectorTask
        _CollectorTask = None


def subdir(name):
    global _TempDirPath
    if _TempDirPath is None:
        init()
    return os.path.join(_TempDirPath, name)


def register(filepath):
    global _FilesDict
    subdir, filename = os.path.split(filepath)
    name = os.path.basename(subdir)
    if name not in _FilesDict.keys():
        name = 'other'
    _FilesDict[name][filepath] = time.time()


# make a new file under sub folder name
# return it's file descriptor and path
# remember you need to close the file descriptor. by your self
# tmpfile will remove it later - do not worry.
def make(name, extension='', prefix=''):
    global _TempDirPath
    global _FilesDict
    if _TempDirPath is None:
        init()
    if name not in _FilesDict.keys():
        name = 'other'
    try:
        fd, filename = tempfile.mkstemp(extension, prefix, subdir(name))
        _FilesDict[name][filename] = time.time()
    except:
        dhnio.Dprint(1, 'tmpfile.make ERROR creating file in sub folder ' + name)
        dhnio.DprintException()
        return None, ''
    dhnio.Dprint(12, 'tmpfile.make ' + filename)
    return fd, filename


def erase(name, filename, why='no reason'):
    global _FilesDict
    if name in _FilesDict.keys():
        try:
            _FilesDict[name].pop(filename, '')
        except:
            dhnio.Dprint(4, 'tmpfile.erase WARNING we do not know about file %s in sub folder %s' %(filename, name))
    else:
        dhnio.Dprint(4, 'tmpfile.erase WARNING we do not know sub folder ' + name)

    if not os.path.exists(filename):
        dhnio.Dprint(6, 'tmpfile.erase WARNING %s not exist' % filename)
        return

    if not os.access(filename, os.W_OK):
        dhnio.Dprint(4, 'tmpfile.erase WARNING %s no write permissions' % filename)
        return

    try:
        os.remove(filename)
        dhnio.Dprint(12, 'tmpfile.erase [%s] deleted because "%s"' % (filename, why))
    except:
        dhnio.Dprint(2, 'tmpfile.erase ERROR can not remove [%s], we tried because %s' % (filename, why))
        #dhnio.DprintException()


def throw_out(filepath, why='dont know'):
    global _FilesDict
    global _SubDirs
    subdir, filename = os.path.split(filepath)
    name = os.path.basename(subdir)
    erase(name, filepath, why)


def collect():
    # remove old files
    # dhnio.Dprint(10, 'tmpfile.collect')
    global _FilesDict
    global _SubDirs
    erase_list = []
    for name in _FilesDict.keys():
        # for data and parity files we have special rules
        # we do not want to remove Data or Parity here.
        # backup_monitor should take care of this.
        if name == 'data-par':
            continue

        else:
            # how long we want to keep the file?
            lifetime = _SubDirs.get(name, 0)
            # if this is not set - keep forever
            if lifetime == 0:
                continue
            for filename, filetime in _FilesDict[name].items():
                # if file is too old - remove it
                if time.time() - filetime > lifetime:
                    erase_list.append((name, filename))

    for name, filename in erase_list:
        erase(name, filename, 'collected')

    dhnio.Dprint(6, 'tmpfile.collect %d files erased' % len(erase_list))

    del erase_list

#    # small fix to solve problem with temp files in the old location
#    try:
#        old_path = os.path.join(os.path.expanduser('~'), 'datahavennet','temp')
#        if os.path.exists(old_path):
#            for filename in os.listdir(old_path):
#                filepath = os.path.join(old_path, filename)
#                try:
#                    os.remove(filepath)
#                    dhnio.Dprint(12, 'tmpfile.erase [%s] deleted because "%s"' % (filepath, 'old location'))
#                except:
#                    dhnio.Dprint(2, 'tmpfile.erase ERROR can not remove ' + filepath)
#    except:
#        dhnio.DprintException()



def startup_clean():
    # at startup we want to scan all sub folders
    # and remove the old files
    # we will get creation time with os.stat
    global _TempDirPath
    global _SubDirs
    dhnio.Dprint(6, 'tmpfile.startup_clean in %s' % _TempDirPath)
    if _TempDirPath is None:
        return
    counter = 0
    for name in os.listdir(_TempDirPath):
        # we want to scan only our folders
        # do not want to be responsible of other files
        if name not in _SubDirs.keys():
            continue

        # for data and parity files we have special rules
        # we do not want to remove Data or Parity here.
        # backup_monitor should take care of this.
        if name == 'data-par':
            pass

        else:
            lifetime = _SubDirs.get(name, 0)
            if lifetime == 0:
                continue
            for filename in os.listdir(subdir(name)):
                filepath = os.path.join(subdir(name), filename)
                filetime = os.stat(filepath).st_ctime
                if time.time() - filetime > lifetime:
                    erase(name, filepath, 'startup cleaning')
                    counter += 1
                    if counter > 200:
                        break

#    # small fix to solve problem with temp files in the old location
#    try:
#        old_path = os.path.join(os.path.expanduser('~'), 'datahavennet','temp')
#        if os.path.exists(old_path):
#            for filename in os.listdir(old_path):
#                filepath = os.path.join(old_path, filename)
#                try:
#                    os.remove(filepath)
#                    dhnio.Dprint(12, 'tmpfile.erase [%s] deleted because "%s"' % (filepath, 'old location'))
#                except:
#                    dhnio.Dprint(2, 'tmpfile.erase ERROR can not remove ' + filepath)
#    except:
#        dhnio.DprintException()

#------------------------------------------------------------------------------


if __name__ == '__main__':
    dhnio.SetDebug(18)
    init()
    fd, filename = make('raid')
    os.write(fd, 'TEST FILE')
    os.close(fd)
    from twisted.internet import reactor
    reactor.run()







#!/usr/bin/python
#backup_matrix.py
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#

import os
import sys
import time
import random
import gc
import cStringIO


try:
    from twisted.internet import reactor
except:
    sys.exit('Error initializing twisted.internet.reactor in backup_matrix.py')

from twisted.internet.defer import Deferred, maybeDeferred
from twisted.internet import threads


import lib.dhnio as dhnio
import lib.misc as misc
import lib.nameurl as nameurl
import lib.transport_control as transport_control
import lib.settings as settings
import lib.contacts as contacts
# import lib.eccmap as eccmap
import lib.tmpfile as tmpfile
import lib.diskspace as diskspace
import lib.packetid as packetid
from lib.automat import Automat


import backup_monitor
import backup_rebuilder  
import fire_hire
import list_files_orator
import backup_db_keeper

import p2p_service
import io_throttle
import contact_status
import backup_fs
import backup_control

#------------------------------------------------------------------------------ 

_RemoteFiles = {}
_LocalFiles = {}
_RemoteMaxBlockNumbers = {}
_LocalMaxBlockNumbers = {}
_LocalBackupSize = {}
_BackupsInProcess = []
_SuppliersSet = None
_BackupStatusNotifyCallback = None
_StatusCallBackForGuiBackup = None
_LocalFilesNotifyCallback = None
_UpdatedBackupIDs = set()
_RepaintingTask = None   
_RepaintingTaskDelay = 2.0

#------------------------------------------------------------------------------ 

def init():
    dhnio.Dprint(4, 'backup_matrix.init')
    RepaintingProcess(True)
    ReadLocalFiles()
    ReadLatestRawListFiles()


def shutdown():
    dhnio.Dprint(4, 'backup_matrix.shutdown')
    RepaintingProcess(False)

#------------------------------------------------------------------------------ 

# index for all remote files (on suppliers HDD's) stored in dictionary
# values are -1, 0 or 1 - this mean file is missing, no info yet, or its existed 
# all entries are indexed by [backupID][blockNum]['D' or 'P'][supplierNum]
def remote_files():
    global _RemoteFiles
    return _RemoteFiles


# max block number for every remote backup ID
def remote_max_block_numbers():
    global _RemoteMaxBlockNumbers
    return _RemoteMaxBlockNumbers


# index for all local files (on our HDD) stored in dictionary
# values are 0 or 1 - this mean file exist or not 
# all entries are indexed by [backupID][blockNum]['D' or 'P'][supplierNum]
def local_files():
    global _LocalFiles
    return _LocalFiles


# max block number for every local backup ID
def local_max_block_numbers():
    global _LocalMaxBlockNumbers
    return _LocalMaxBlockNumbers


def local_backup_size():
    global _LocalBackupSize
    return _LocalBackupSize


# currently working backups (started from dobackup.py)
def backups_in_process():
    global _BackupsInProcess
    return _BackupsInProcess


def suppliers_set():
    global _SuppliersSet
    if _SuppliersSet is None:
        _SuppliersSet = SuppliersSet(contacts.getSupplierIDs())
    return _SuppliersSet

#------------------------------------------------------------------------------ 
# this should represent the set of suppliers for either the user or for a customer the user
# is acting as a scrubber for
class SuppliersSet:
    def __init__(self, supplierList):
        self.suppliers = [] 
        self.supplierCount = 0 
        self.UpdateSuppliers(supplierList)

    def GetActiveArray(self):
        activeArray = [0] * self.supplierCount
        for i in range(self.supplierCount):
            if not self.suppliers[i]:
                continue
            if contact_status.isOnline(self.suppliers[i]):
                activeArray[i] = 1
            else:
                activeArray[i] = 0
        return activeArray

    def ChangedArray(self, supplierList):
        changedArray = [0] * self.supplierCount
        changedIdentities = []
        for i in range(self.supplierCount):
            if not self.suppliers[i]:
                continue
            if supplierList[i] != self.suppliers[i]:
                changedArray[i] = 1
                changedIdentities.append(supplierList[i])
        return changedArray, changedIdentities

    def SuppliersChanged(self, supplierList):
        if len(supplierList) != self.supplierCount:
            return True
        for i in range(self.supplierCount):
            if not self.suppliers[i]:
                continue
            if supplierList[i] != self.suppliers[i]:
                return True
        return False

    # if suppliers 1 and 3 changed, return [1,3]
    def SuppliersChangedNumbers(self, supplierList):
        changedList = []
        for i in range(self.supplierCount):
            if not self.suppliers[i]:
                continue
            if supplierList[i] != self.suppliers[i]:
                changedList.append(i)
        return changedList

    def SupplierCountChanged(self, supplierList):
        if len(supplierList) != self.supplierCount:
            return True
        else:
            return False

    def UpdateSuppliers(self, supplierList):
        self.suppliers = supplierList
        self.supplierCount = len(self.suppliers)

#------------------------------------------------------------------------------ 

def SaveLatestRawListFiles(idurl, listFileText):
    supplierPath = settings.SupplierPath(idurl)
    if not os.path.isdir(supplierPath):
        try:
            os.makedirs(supplierPath)
        except:
            dhnio.DprintException()
            return
    dhnio.WriteFile(settings.SupplierListFilesFilename(idurl), listFileText)


def ReadRawListFiles(supplierNum, listFileText):
    # all lines are something like that
    # Findex
    # D0
    # D0/1 
    # V0/1/F20090709034221PM 3 0-1000
    # V0/1/F20090709034221PM 3 0-1000
    # D0/0/123/4567
    # V0/0/123/4567/F20090709034221PM 3 0-11 missing Data:1,3
    # V0/0/123/4/F20090709012331PM 3 0-5 missing Data:1,3 Parity:0,1,2
    backups2remove = set()
    paths2remove = set()
    oldfiles = ClearSupplierRemoteInfo(supplierNum)
    newfiles = 0
    dhnio.Dprint(8, 'backup_matrix.ReadRawListFiles %d bytes to read' % len(listFileText))
    input = cStringIO.StringIO(listFileText)
    while True:
        line = input.readline()
        if line == '':
            break
        typ = line[0]
        line = line[1:] 
        line = line.rstrip('\n')
        # comment lines in Files reports start with blank,
        if line.strip() == '':
            continue
        # also don't consider the identity a backup,
        if line.find('http://') != -1 or line.find('.xml') != -1:
            continue
        # nor backup_info.xml, nor backup_db, nor index  files 
        if line in [ settings.BackupIndexFileName(), settings.BackupInfoFileName(), settings.BackupInfoFileNameOld(), settings.BackupInfoEncryptedFileName() ]:
            continue
        if typ == 'F':
            # we don't have this path in the index
            # so we have several cases:
            #    1. this is old file and we need to remove it and all its backups
            #    2. we loose our local index and did not restore it from one of suppliers yet
            #    3. we did restore our account and did not restore the index yet
            #    4. we lost our index at all and we do not have nor local nor remote copy
            # what to do now: 
            #    - in first case we just need to remove the file from remote supplier
            #    - in other cases we must keep all remote data and believe we can restore the index 
            #         and get all file names and backed up data
            # how to recognize that? how to be sure we have the correct index?
            # because it should be empty right after we recover our account 
            # or we may loose it if the local index file were lost
            # the first idea:  check backup_db_keeper state - READY means index is fine
            # the second idea: check revision number of the local index - 0 means we have no index yet 
            if not backup_fs.IsFileID(line): # remote supplier have some file - but we don't have it in the index
                if backup_control.revision() > 0 and backup_db_keeper.A().IsSynchronized():  
                    # so we have some modifications in the index - it is not empty!
                    # backup_db_keeper did its job - so we have the correct index
                    paths2remove.add(line) # now we are sure that this file is old and must be removed
                    dhnio.Dprint(8, '        F%s - remove, not found in the index' % line)
                # what to do now? let's hope we still can restore our index and this file is our remote data
        elif typ == 'D':
            if not backup_fs.ExistsID(line):
                if backup_control.revision() > 0 and backup_db_keeper.A().IsSynchronized():
                    paths2remove.add(line)
                    dhnio.Dprint(8, '        D%s - remove, not found in the index' % line)
        elif typ == 'V':
            words = line.split(' ')
            # minimum is 3 words: "0/0/F20090709034221PM", "3", "0-1000"
            if len(words) < 3:
                dhnio.Dprint(2, 'backup_matrix.ReadRawListFiles WARNING incorrect line:[%s]' % line)
                continue
            try:
                pathID, versionName = packetid.SplitBackupID(words[0])
                backupID = pathID+'/'+versionName
                lineSupplierNum = int(words[1])
                minBlockNum, maxBlockNum = words[2].split('-')
                maxBlockNum = int(maxBlockNum)
            except:
                dhnio.Dprint(2, 'backup_matrix.ReadRawListFiles WARNING incorrect line:[%s]' % line)
                continue
            if lineSupplierNum != supplierNum:
                # this mean supplier have old files and we do not need that 
                backups2remove.add(backupID)
                dhnio.Dprint(8, '        V%s - remove, different supplier number' % backupID)
                continue
            iter_path = backup_fs.WalkByID(pathID)
            if iter_path is None:
                # this version is not found in the index
                if backup_control.revision() > 0 and backup_db_keeper.A().IsSynchronized():
                    backups2remove.add(backupID)
                    paths2remove.add(pathID)
                    dhnio.Dprint(8, '        V%s - remove, path not found in the index' % pathID)
                continue
            item, localPath = iter_path
            if isinstance(item, dict):
                try:
                    item = item[backup_fs.INFO_KEY]
                except:
                    item = None
            if not item or not item.has_version(versionName):
                if backup_control.revision() > 0 and backup_db_keeper.A().IsSynchronized():
                    backups2remove.add(backupID)
                    dhnio.Dprint(8, '        V%s - remove, version is not found in the index' % backupID)
                continue
            missingBlocksSet = {'Data': set(), 'Parity': set()}
            if len(words) > 3:
                # "0/0/123/4567/F20090709034221PM/0-Data" "3" "0-5" "missing" "Data:1,3" "Parity:0,1,2"
                if words[3].strip() != 'missing':
                    dhnio.Dprint(2, 'backup_matrix.ReadRawListFiles WARNING incorrect line:[%s]' % line)
                    continue
                for missingBlocksString in words[4:]:
                    try:
                        dp, blocks = missingBlocksString.split(':')
                        missingBlocksSet[dp] = set(blocks.split(','))
                    except:
                        dhnio.DprintException()
                        break
            if not remote_files().has_key(backupID):
                remote_files()[backupID] = {}
                # dhnio.Dprint(6, 'backup_matrix.ReadRawListFiles new remote entry for %s created in the memory' % backupID)
            # +1 because range(2) give us [0,1] but we want [0,1,2]
            for blockNum in range(maxBlockNum+1):
                if not remote_files()[backupID].has_key(blockNum):
                    remote_files()[backupID][blockNum] = {
                        'D': [0] * suppliers_set().supplierCount,
                        'P': [0] * suppliers_set().supplierCount,}
                for dataORparity in ['Data', 'Parity']:
                    # we set -1 if the file is missing and 1 if exist, so 0 mean "no info yet" ... smart!
                    bit = -1 if str(blockNum) in missingBlocksSet[dataORparity] else 1 
                    remote_files()[backupID][blockNum][dataORparity[0]][supplierNum] = bit
                    newfiles += int((bit + 1) / 2) # this should switch -1 or 1 to 0 or 1
            # save max block number for this backup
            if not remote_max_block_numbers().has_key(backupID):
                remote_max_block_numbers()[backupID] = -1 
            if maxBlockNum > remote_max_block_numbers()[backupID]:
                remote_max_block_numbers()[backupID] = maxBlockNum
            # mark this backup to be repainted
            RepaintBackup(backupID)
    input.close()
    dhnio.Dprint(8, 'backup_matrix.ReadRawListFiles for supplier %d, old/new files:%d/%d, backups2remove:%d, paths2remove:%d' % (
        supplierNum, oldfiles, newfiles, len(backups2remove), len(paths2remove)))
    # return list of backupID's which is too old but stored on suppliers machines 
    return backups2remove, paths2remove
            

def ReadLatestRawListFiles():
    dhnio.Dprint(4, 'backup_matrix.ReadLatestRawListFiles')
    for idurl in contacts.getSupplierIDs():
        if idurl:
            filename = os.path.join(settings.SupplierPath(idurl, 'listfiles'))
            if os.path.isfile(filename):
                listFileText = dhnio.ReadTextFile(filename)
                if listFileText.strip() != '':
                    ReadRawListFiles(contacts.numberForSupplier(idurl), listFileText)

def ReadLocalFiles():
    global _LocalFilesNotifyCallback
    local_files().clear()
    local_max_block_numbers().clear()
    local_backup_size().clear()
    _counter = [0,]
    def visit(realpath, subpath, name):
        # subpath is something like 0/0/1/0/F20131120053803PM/0-1-Data  
        if not os.path.isfile(realpath):
            return True
        if realpath.startswith('newblock-'):
            return False
        if subpath in [ settings.BackupIndexFileName(), settings.BackupInfoFileName(), settings.BackupInfoFileNameOld(), settings.BackupInfoEncryptedFileName() ]:
            return False
        try:
            version = subpath.split('/')[-2]
        except:
            return False
        if not packetid.IsCanonicalVersion(version):
            return True
        LocalFileReport(packetID=subpath)
        _counter[0] += 1
        return False
    dhnio.traverse_dir_recursive(visit, settings.getLocalBackupsDir())
    dhnio.Dprint(8, 'backup_matrix.ReadLocalFiles  %d files indexed' % _counter[0])
    if dhnio.Debug(8):
        try:
            if sys.version_info >= (2, 6):
                #localSZ = sys.getsizeof(local_files())
                #remoteSZ = sys.getsizeof(remote_files())
                import lib.getsizeof
                localSZ = lib.getsizeof.total_size(local_files())
                remoteSZ = lib.getsizeof.total_size(remote_files())
                indexByName = lib.getsizeof.total_size(backup_fs.fs())
                indexByID = lib.getsizeof.total_size(backup_fs.fsID())
                dhnio.Dprint(10, '    all local info uses %d bytes in the memory' % localSZ)
                dhnio.Dprint(10, '    all remote info uses %d bytes in the memory' % remoteSZ)
                dhnio.Dprint(10, '    index by name takes %d bytes in the memory' % indexByName)
                dhnio.Dprint(10, '    index by ID takes %d bytes in the memory' % indexByID)
        except:
            dhnio.DprintException()
    if _LocalFilesNotifyCallback is not None:
        _LocalFilesNotifyCallback()

#------------------------------------------------------------------------------ 

def RemoteFileReport(backupID, blockNum, supplierNum, dataORparity, result):
    blockNum = int(blockNum)
    supplierNum = int(supplierNum)
    if supplierNum > suppliers_set().supplierCount:
        dhnio.Dprint(4, 'backup_matrix.BackupReport got too big supplier number, possible this is an old packet')
        return
    if not remote_files().has_key(backupID):
        remote_files()[backupID] = {}
        dhnio.Dprint(8, 'backup_matrix.BackupReport new remote entry for %s created in the memory' % backupID)
    if not remote_files()[backupID].has_key(blockNum):
        remote_files()[backupID][blockNum] = {
            'D': [0] * suppliers_set().supplierCount,
            'P': [0] * suppliers_set().supplierCount,}
    # save backed up block info into remote info structure, synchronize on hand info
    flag = 1 if result else 0
    if dataORparity == 'Data':
        remote_files()[backupID][blockNum]['D'][supplierNum] = flag 
    elif dataORparity == 'Parity':
        remote_files()[backupID][blockNum]['P'][supplierNum] = flag
    else:
        dhnio.Dprint(4, 'backup_matrix.BackupReport WARNING incorrect packet ID: %s' % packetID)
    # if we know only 5 blocks stored on remote machine
    # but we have backed up 6th block - remember this  
    remote_max_block_numbers()[backupID] = max(remote_max_block_numbers().get(backupID, -1), blockNum)
    # mark to repaint this backup in gui
    RepaintBackup(backupID)


def LocalFileReport(packetID=None, backupID=None, blockNum=None, supplierNum=None, dataORparity=None):
    if packetID is not None:
        backupID, blockNum, supplierNum, dataORparity = packetid.Split(packetID)  
        if backupID is None:
            dhnio.Dprint(8, 'backup_matrix.LocalFileReport WARNING incorrect filename: ' + packetID)
            return
    else:
        blockNum = int(blockNum)
        supplierNum = int(supplierNum)
        dataORparity = dataORparity
        packetID = packetid.MakePacketID(backupID, blockNum, supplierNum, dataORparity)
    filename = packetID
    if dataORparity not in ['Data', 'Parity']:
        dhnio.Dprint(4, 'backup_matrix.LocalFileReport WARNING Data or Parity? ' + filename)
        return
    if supplierNum >= suppliers_set().supplierCount:
        dhnio.Dprint(4, 'backup_matrix.LocalFileReport WARNING supplier number %d > %d %s' % (supplierNum, suppliers_set().supplierCount, filename))
        return
    if not local_files().has_key(backupID):
        local_files()[backupID] = {}
        # dhnio.Dprint(14, 'backup_matrix.LocalFileReport new local entry for %s created in the memory' % backupID)
    if not local_files()[backupID].has_key(blockNum):
        local_files()[backupID][blockNum] = {
            'D': [0] * suppliers_set().supplierCount,
            'P': [0] * suppliers_set().supplierCount}
    local_files()[backupID][blockNum][dataORparity[0]][supplierNum] = 1
    if not local_max_block_numbers().has_key(backupID):
        local_max_block_numbers()[backupID] = -1
    if local_max_block_numbers()[backupID] < blockNum:
        local_max_block_numbers()[backupID] = blockNum
    # dhnio.Dprint(6, 'backup_matrix.LocalFileReport %s max block num is %d' % (backupID, local_max_block_numbers()[backupID]))
    if not local_backup_size().has_key(backupID):
        local_backup_size()[backupID] = 0
    try:
        local_backup_size()[backupID] += os.path.getsize(os.path.join(settings.getLocalBackupsDir(), filename))
    except:
        dhnio.DprintException()
    RepaintBackup(backupID)


def LocalBlockReport(newblock, num_suppliers):
    if suppliers_set().supplierCount != num_suppliers:
        dhnio.Dprint(6, 'backup_matrix.LocalBlockReport %s skipped, because number of suppliers were changed' % str(newblock))
        return
    try:
        backupID = newblock.BackupID
        blockNum = int(newblock.BlockNumber)
    except:
        dhnio.DprintException()
        return
    for supplierNum in range(num_suppliers):
        for dataORparity in ('Data', 'Parity'):
            packetID = packetid.MakePacketID(backupID, blockNum, supplierNum, dataORparity)
            if not local_files().has_key(backupID):
                local_files()[backupID] = {}
                # dhnio.Dprint(14, 'backup_matrix.LocalFileReport new local entry for %s created in the memory' % backupID)
            if not local_files()[backupID].has_key(blockNum):
                local_files()[backupID][blockNum] = {
                    'D': [0] * suppliers_set().supplierCount,
                    'P': [0] * suppliers_set().supplierCount}
            local_files()[backupID][blockNum][dataORparity[0]][supplierNum] = 1
            # dhnio.Dprint(6, 'backup_matrix.LocalFileReport %s max block num is %d' % (backupID, local_max_block_numbers()[backupID]))
            if not local_backup_size().has_key(backupID):
                local_backup_size()[backupID] = 0
            try:
                local_backup_size()[backupID] += os.path.getsize(os.path.join(settings.getLocalBackupsDir(), packetID))
            except:
                dhnio.DprintException()
    if not local_max_block_numbers().has_key(backupID):
        local_max_block_numbers()[backupID] = -1
    if local_max_block_numbers()[backupID] < blockNum:
        local_max_block_numbers()[backupID] = blockNum
    RepaintBackup(backupID)

#------------------------------------------------------------------------------ 

def ScanMissingBlocks(backupID):
    missingBlocks = set()
    localMaxBlockNum = local_max_block_numbers().get(backupID, -1)
    remoteMaxBlockNum = remote_max_block_numbers().get(backupID, -1)
    supplierActiveArray = suppliers_set().GetActiveArray()

    if not remote_files().has_key(backupID):
        if not local_files().has_key(backupID):
            # we have no local and no remote info for this backup
            # no chance to do some rebuilds...
            # TODO but how we get here ?! 
            dhnio.Dprint(4, 'backup_matrix.ScanMissingBlocks no local and no remote info for %s' % backupID)
        else:
            # we have no remote info, but some local files exists
            # so let's try to sent all of them
            # need to scan all block numbers 
            for blockNum in range(localMaxBlockNum):
                # we check for Data and Parity packets
                localData = GetLocalDataArray(backupID, blockNum)
                localParity = GetLocalParityArray(backupID, blockNum)  
                for supplierNum in range(len(supplierActiveArray)):
                    # if supplier is not alive we can not send to him
                    # so no need to scan for missing blocks 
                    if supplierActiveArray[supplierNum] != 1:
                        continue
                    if localData[supplierNum] == 1:
                        missingBlocks.add(blockNum)
                    if localParity[supplierNum] == 1:
                        missingBlocks.add(blockNum)
    else:
        # now we have some remote info
        # we take max block number from local and remote
        maxBlockNum = max(remoteMaxBlockNum, localMaxBlockNum)
        # dhnio.Dprint(6, 'backup_matrix.ScanMissingBlocks maxBlockNum=%d' % maxBlockNum)
        # and increase by one because range(3) give us [0, 1, 2], but we want [0, 1, 2, 3]
        for blockNum in range(maxBlockNum + 1):
            # if we have few remote files, but many locals - we want to send all missed 
            if not remote_files()[backupID].has_key(blockNum):
                missingBlocks.add(blockNum)
                continue
            # take remote info for this block
            remoteData = GetRemoteDataArray(backupID, blockNum)
            remoteParity = GetRemoteParityArray(backupID, blockNum)  
            # now check every our supplier for every block
            for supplierNum in range(len(supplierActiveArray)):
                # if supplier is not alive we can not send to him
                # so no need to scan for missing blocks 
                if supplierActiveArray[supplierNum] != 1:
                    continue
                if remoteData[supplierNum] != 1:    # -1 means missing
                    missingBlocks.add(blockNum)     # 0 - no info yet
                if remoteParity[supplierNum] != 1:  # 1 - file exist on remote supplier 
                    missingBlocks.add(blockNum)
                
    # dhnio.Dprint(6, 'backup_matrix.ScanMissingBlocks %s' % missingBlocks)
    return list(missingBlocks)

def ScanBlocksToRemove(backupID, check_all_suppliers=True):
    dhnio.Dprint(10, 'backup_matrix.ScanBlocksToRemove for %s' % backupID)
    packets = []
    localMaxBlockNum = local_max_block_numbers().get(backupID, -1)
    if not remote_files().has_key(backupID) or not local_files().has_key(backupID):
        # no info about this backup yet - skip
        return packets
    for blockNum in range(localMaxBlockNum + 1):
        localArray = {'Data': GetLocalDataArray(backupID, blockNum),
                      'Parity': GetLocalParityArray(backupID, blockNum)}  
        remoteArray = {'Data': GetRemoteDataArray(backupID, blockNum),
                       'Parity': GetRemoteParityArray(backupID, blockNum)}  
        if ( 0 in remoteArray['Data'] ) or ( 0 in remoteArray['Parity'] ):
            # if some supplier do not have some data for that block - do not remove any local files for that block!
            # we do remove the local files only when we sure all suppliers got the all data pieces
            continue
        if ( -1 in remoteArray['Data'] ) or ( -1 in remoteArray['Parity'] ):
            # also if we do not have any info about this block for some supplier do not remove other local pieces
            continue
        for supplierNum in range(suppliers_set().supplierCount):
            supplierIDURL = suppliers_set().suppliers[supplierNum]
            if not supplierIDURL:
                # supplier is unknown - skip
                continue
            for dataORparity in ['Data', 'Parity']:
                packetID = packetid.MakePacketID(backupID, blockNum, supplierNum, dataORparity)
                if io_throttle.HasPacketInSendQueue(supplierIDURL, packetID):
                    # if we do sending the packet at the moment - skip
                    continue
                if localArray[dataORparity][supplierNum] == 1:  
                    packets.append(packetID)
                    dhnio.Dprint(10, '    mark to remove %s, blockNum:%d remote:%s local:%s' % (packetID, blockNum, str(remoteArray), str(localArray)))
#                if check_all_suppliers:
#                    if localArray[dataORparity][supplierNum] == 1:  
#                        packets.append(packetID)
#                else:
#                    if remoteArray[dataORparity][supplierNum] == 1 and localArray[dataORparity][supplierNum] == 1:  
#                        packets.append(packetID)
    return packets

def ScanBlocksToSend(backupID):
    if '' in suppliers_set().suppliers:
        return {} 
    localMaxBlockNum = local_max_block_numbers().get(backupID, -1)
    supplierActiveArray = suppliers_set().GetActiveArray()
    bySupplier = {}
    for supplierNum in range(len(supplierActiveArray)):
        bySupplier[supplierNum] = set()
    if not remote_files().has_key(backupID):
        for blockNum in range(localMaxBlockNum + 1):
            localData = GetLocalDataArray(backupID, blockNum)
            localParity = GetLocalParityArray(backupID, blockNum)  
            for supplierNum in range(len(supplierActiveArray)):
                if supplierActiveArray[supplierNum] != 1:
                    continue
                if localData[supplierNum] == 1:
                    bySupplier[supplierNum].add(packetid.MakePacketID(backupID, blockNum, supplierNum, 'Data'))
                if localParity[supplierNum] == 1:
                    bySupplier[supplierNum].add(packetid.MakePacketID(backupID, blockNum, supplierNum, 'Parity'))
    else:
        for blockNum in range(localMaxBlockNum + 1):
            remoteData = GetRemoteDataArray(backupID, blockNum)
            remoteParity = GetRemoteParityArray(backupID, blockNum)  
            localData = GetLocalDataArray(backupID, blockNum)
            localParity = GetLocalParityArray(backupID, blockNum)  
            for supplierNum in range(len(supplierActiveArray)):
                if supplierActiveArray[supplierNum] != 1:
                    continue
                if remoteData[supplierNum] != 1 and localData[supplierNum] == 1:    
                    bySupplier[supplierNum].add(packetid.MakePacketID(backupID, blockNum, supplierNum, 'Data'))   
                if remoteParity[supplierNum] != 1 and localParity[supplierNum] == 1:   
                    bySupplier[supplierNum].add(packetid.MakePacketID(backupID, blockNum, supplierNum, 'Parity'))
    return bySupplier

#------------------------------------------------------------------------------ 

def RepaintBackup(backupID): 
    global _UpdatedBackupIDs
    _UpdatedBackupIDs.add(backupID)


def RepaintingProcess(on_off):
    global _UpdatedBackupIDs
    global _BackupStatusNotifyCallback
    global _RepaintingTask
    global _RepaintingTaskDelay
    if on_off is False:
        _RepaintingTaskDelay = 2.0
        if _RepaintingTask is not None:
            if _RepaintingTask.active():
                _RepaintingTask.cancel()
                _RepaintingTask = None
                _UpdatedBackupIDs.clear()
                return
    for backupID in _UpdatedBackupIDs:
        if _BackupStatusNotifyCallback is not None:
            _BackupStatusNotifyCallback(backupID)
    _RepaintingTaskDelay = misc.LoopAttenuation(_RepaintingTaskDelay, len(_UpdatedBackupIDs) > 0, 2.0, 8.0)
    _UpdatedBackupIDs.clear()
#        _RepaintingTaskDelay = 0.5
#    else:
#        if _RepaintingTaskDelay < 2.0:
#            _RepaintingTaskDelay *= 2.0
    # attenuation
    _RepaintingTask = reactor.callLater(_RepaintingTaskDelay, RepaintingProcess, True)

#------------------------------------------------------------------------------ 

def EraseBackupRemoteInfo(backupID): 
    if remote_files().has_key(backupID):
        del remote_files()[backupID] # remote_files().pop(backupID)
    if remote_max_block_numbers().has_key(backupID):
        del remote_max_block_numbers()[backupID]
        
def EraseBackupLocalInfo(backupID):
    if local_files().has_key(backupID):
        del local_files()[backupID] # local_files().pop(backupID)
    if local_max_block_numbers().has_key(backupID):
        del local_max_block_numbers()[backupID]
    if local_backup_size().has_key(backupID):
        del local_backup_size()[backupID]

#------------------------------------------------------------------------------ 

def ClearLocalInfo():
    local_files().clear()
    local_max_block_numbers().clear()
    local_backup_size().clear()

def ClearRemoteInfo():
    remote_files().clear()
    remote_max_block_numbers().clear()
    
def ClearSupplierRemoteInfo(supplierNum):
    files = 0
    for backupID in remote_files().keys():
        for blockNum in remote_files()[backupID].keys():
            if remote_files()[backupID][blockNum]['D'][supplierNum] == 1:
                files += 1 
            if remote_files()[backupID][blockNum]['P'][supplierNum] == 1:
                files += 1
            remote_files()[backupID][blockNum]['D'][supplierNum] = 0
            remote_files()[backupID][blockNum]['P'][supplierNum] = 0
    return files

#------------------------------------------------------------------------------ 

def GetBackupStats(backupID):
    if not remote_files().has_key(backupID):
        return 0, 0, [(0, 0)] * suppliers_set().supplierCount
    percentPerSupplier = 100.0 / suppliers_set().supplierCount
    # ??? maxBlockNum = remote_max_block_numbers().get(backupID, -1)
    maxBlockNum = GetKnownMaxBlockNum(backupID)
    fileNumbers = [0] * suppliers_set().supplierCount
    totalNumberOfFiles = 0
    for blockNum in remote_files()[backupID].keys():
        for supplierNum in range(len(fileNumbers)):
            if supplierNum < suppliers_set().supplierCount:
                if remote_files()[backupID][blockNum]['D'][supplierNum] == 1:
                    fileNumbers[supplierNum] += 1
                    totalNumberOfFiles += 1
                if remote_files()[backupID][blockNum]['P'][supplierNum] == 1:
                    fileNumbers[supplierNum] += 1
                    totalNumberOfFiles += 1
    statsArray = []
    for supplierNum in range(suppliers_set().supplierCount):
        if maxBlockNum > -1:
            # 0.5 because we count both Parity and Data.
            percent = percentPerSupplier * 0.5 * fileNumbers[supplierNum] / ( maxBlockNum + 1 )
        else:
            percent = 0.0
        statsArray.append(( percent, fileNumbers[supplierNum] ))
    del fileNumbers 
    return totalNumberOfFiles, maxBlockNum, statsArray


# return totalPercent, totalNumberOfFiles, totalSize, maxBlockNum, statsArray
def GetBackupLocalStats(backupID):
    # ??? maxBlockNum = local_max_block_numbers().get(backupID, -1)
    maxBlockNum = GetKnownMaxBlockNum(backupID)
    if not local_files().has_key(backupID):
        return 0, 0, 0, maxBlockNum, [(0, 0)] * suppliers_set().supplierCount
    percentPerSupplier = 100.0 / suppliers_set().supplierCount
    totalNumberOfFiles = 0
    fileNumbers = [0] * suppliers_set().supplierCount
    for blockNum in range(maxBlockNum + 1):
        if blockNum not in local_files()[backupID].keys():
            continue
#    for blockNum in local_files()[backupID].keys():
        for supplierNum in range(len(fileNumbers)):
            if supplierNum < suppliers_set().supplierCount:
                if local_files()[backupID][blockNum]['D'][supplierNum] == 1:
                    fileNumbers[supplierNum] += 1
                    totalNumberOfFiles += 1
                if local_files()[backupID][blockNum]['P'][supplierNum] == 1:
                    fileNumbers[supplierNum] += 1
                    totalNumberOfFiles += 1
    statsArray = []
    for supplierNum in range(suppliers_set().supplierCount):
        if maxBlockNum > -1:
            # 0.5 because we count both Parity and Data.
            percent = percentPerSupplier * 0.5 * fileNumbers[supplierNum] / ( maxBlockNum + 1 )
        else:
            percent = 0.0
        statsArray.append(( percent, fileNumbers[supplierNum] ))
    del fileNumbers 
    totalPercent = 100.0 * 0.5 * totalNumberOfFiles / ((maxBlockNum + 1) * suppliers_set().supplierCount)
    return totalPercent, totalNumberOfFiles, local_backup_size().get(backupID, 0), maxBlockNum, statsArray


def GetBackupBlocksAndPercent(backupID):
    if not remote_files().has_key(backupID):
        return 0, 0
    # get max block number
    # ??? maxBlockNum = remote_max_block_numbers().get(backupID, -1)
    maxBlockNum = GetKnownMaxBlockNum(backupID)
    if maxBlockNum == -1:
        return 0, 0
    # we count all remote files for this backup
    fileCounter = 0
    for blockNum in remote_files()[backupID].keys():
        for supplierNum in range(suppliers_set().supplierCount):
            if remote_files()[backupID][blockNum]['D'][supplierNum] == 1:
                fileCounter += 1
            if remote_files()[backupID][blockNum]['P'][supplierNum] == 1:
                fileCounter += 1
    # +1 since zero based and *0.5 because Data and Parity
    return maxBlockNum + 1, 100.0 * 0.5 * fileCounter / ((maxBlockNum + 1) * suppliers_set().supplierCount)

# return : blocks, percent, weak block, weak percent
def GetBackupRemoteStats(backupID, only_available_files=True):
    if not remote_files().has_key(backupID):
        return 0, 0, 0, 0
    # get max block number
    # ??? maxBlockNum = remote_max_block_numbers().get(backupID, -1)
    maxBlockNum = GetKnownMaxBlockNum(backupID)
    if maxBlockNum == -1:
        return 0, 0, 0, 0
    supplierCount = suppliers_set().supplierCount
    fileCounter = 0
    weakBlockNum = -1
    lessSuppliers = supplierCount
    activeArray = suppliers_set().GetActiveArray()
    # we count all remote files for this backup - scan all blocks
    for blockNum in range(maxBlockNum + 1):
        if blockNum not in remote_files()[backupID].keys():
            lessSuppliers = 0
            weakBlockNum = blockNum
            continue
        goodSuppliers = supplierCount
        for supplierNum in range(supplierCount):
            if activeArray[supplierNum] != 1 and only_available_files:
                goodSuppliers -= 1
                continue
            if remote_files()[backupID][blockNum]['D'][supplierNum] != 1 or remote_files()[backupID][blockNum]['P'][supplierNum] != 1:
                goodSuppliers -= 1
            if remote_files()[backupID][blockNum]['D'][supplierNum] == 1:
                fileCounter += 1
            if remote_files()[backupID][blockNum]['P'][supplierNum] == 1:
                fileCounter += 1
        if goodSuppliers < lessSuppliers:
            lessSuppliers = goodSuppliers
            weakBlockNum = blockNum
    # +1 since zero based and *0.5 because Data and Parity
    return (maxBlockNum + 1, 100.0 * 0.5 * fileCounter / ((maxBlockNum + 1) * supplierCount),
            weakBlockNum, 100.0 * float(lessSuppliers) / float(supplierCount))


def GetBackupRemoteArray(backupID):
    if not remote_files().has_key(backupID):
        return None
    maxBlockNum = GetKnownMaxBlockNum(backupID)
    if maxBlockNum == -1:
        return None
    return remote_files()[backupID]


def GetBackupLocalArray(backupID):
    if not local_files().has_key(backupID):
        return None
    maxBlockNum = GetKnownMaxBlockNum(backupID)
    if maxBlockNum == -1:
        return None
    return local_files()[backupID]
        
    
def GetBackupIDs(remote=True, local=False, sorted_ids=False):
    s = set()
    if remote:
        s.update(remote_files().keys())
    if local:
        s.update(local_files().keys())
    if sorted_ids:
        return misc.sorted_backup_ids(list(s))
    return list(s)


def GetKnownMaxBlockNum(backupID):
    return max(remote_max_block_numbers().get(backupID, -1), 
               local_max_block_numbers().get(backupID, -1))


def GetLocalDataArray(backupID, blockNum):
    if not local_files().has_key(backupID):
        return [0] * suppliers_set().supplierCount
    if not local_files()[backupID].has_key(blockNum):
        return [0] * suppliers_set().supplierCount
    return local_files()[backupID][blockNum]['D']


def GetLocalParityArray(backupID, blockNum):
    if not local_files().has_key(backupID):
        return [0] * suppliers_set().supplierCount
    if not local_files()[backupID].has_key(blockNum):
        return [0] * suppliers_set().supplierCount
    return local_files()[backupID][blockNum]['P']
    

def GetRemoteDataArray(backupID, blockNum):
    if not remote_files().has_key(backupID):
        return [0] * suppliers_set().supplierCount
    if not remote_files()[backupID].has_key(blockNum):
        return [0] * suppliers_set().supplierCount
    return remote_files()[backupID][blockNum]['D']

    
def GetRemoteParityArray(backupID, blockNum):
    if not remote_files().has_key(backupID):
        return [0] * suppliers_set().supplierCount
    if not remote_files()[backupID].has_key(blockNum):
        return [0] * suppliers_set().supplierCount
    return remote_files()[backupID][blockNum]['P']


def GetSupplierStats(supplierNum):
    result = {}
    files = total = 0
    for backupID in remote_files().keys():
        result[backupID] = [0, 0]
        for blockNum in remote_files()[backupID].keys():
            if remote_files()[backupID][blockNum]['D'][supplierNum] == 1:
                result[backupID][0] += 1
                files += 1 
            if remote_files()[backupID][blockNum]['P'][supplierNum] == 1:
                result[backupID][0] += 1
                files += 1
            result[backupID][1] += 2
            total += 2
    return files, total, result


def GetWeakLocalBlock(backupID):
    # scan all local blocks for this backup 
    # and find the worst block 
    supplierCount = suppliers_set().supplierCount
    if not local_files().has_key(backupID):
        return -1, 0, supplierCount
    maxBlockNum = GetKnownMaxBlockNum(backupID)
    weakBlockNum = -1
    lessSuppliers = supplierCount
    for blockNum in range(maxBlockNum+1):
        if blockNum not in local_files()[backupID].keys():
            return blockNum, 0, supplierCount
        goodSuppliers = supplierCount
        for supplierNum in range(supplierCount):
            if  local_files()[backupID][blockNum]['D'][supplierNum] != 1 or local_files()[backupID][blockNum]['P'][supplierNum] != 1:
                goodSuppliers -= 1
        if goodSuppliers < lessSuppliers:
            lessSuppliers = goodSuppliers
            weakBlockNum = blockNum
    return weakBlockNum, lessSuppliers, supplierCount
 
    
def GetWeakRemoteBlock(backupID):
    # scan all remote blocks for this backup 
    # and find the worst block - less suppliers keeps the data and stay online 
    supplierCount = suppliers_set().supplierCount
    if not remote_files().has_key(backupID):
        return -1, 0, supplierCount
    maxBlockNum = GetKnownMaxBlockNum(backupID)
    weakBlockNum = -1
    lessSuppliers = supplierCount
    activeArray = suppliers_set().GetActiveArray()
    for blockNum in range(maxBlockNum+1):
        if blockNum not in remote_files()[backupID].keys():
            return blockNum, 0, supplierCount
        goodSuppliers = supplierCount
        for supplierNum in range(supplierCount):
            if activeArray[supplierNum] != 1:
                goodSuppliers -= 1
                continue
            if  remote_files()[backupID][blockNum]['D'][supplierNum] != 1 or remote_files()[backupID][blockNum]['P'][supplierNum] != 1:
                goodSuppliers -= 1
        if goodSuppliers < lessSuppliers:
            lessSuppliers = goodSuppliers
            weakBlockNum = blockNum
    return weakBlockNum, lessSuppliers, supplierCount

#------------------------------------------------------------------------------ 

def SetBackupStatusNotifyCallback(callBack):
    global _BackupStatusNotifyCallback
    _BackupStatusNotifyCallback = callBack

def SetLocalFilesNotifyCallback(callback):
    global _LocalFilesNotifyCallback
    _LocalFilesNotifyCallback = callback

#------------------------------------------------------------------------------ 


if __name__ == "__main__":
    init()
    import pprint
    pprint.pprint(GetBackupIds())














from random import randint, choice
from time import time
from bisect import bisect_left, insort_right
from nitro.async import AsyncOp
from cspace.dht.params import DHT_ID_MAX
from cspace.dht.util import numToId
import logging

logger = logging.getLogger('cspace.network.nodetable')

MAX_NODES = 100
NODETABLE_PING_INTERVAL = 1*60
NODETABLE_LOOKUP_INTERVAL = 6*60

FIND_LIVE_NODES_TIMEOUT = 3
FIND_LIVE_NODES_TIMEOUT_COUNT = 1

class _NodeInfo( object ) :
    def __init__( self, nodeAddr, isSeedNode, sortId ) :
        self.nodeAddr = nodeAddr
        self.isSeedNode = isSeedNode
        self.sortId = sortId
        self.firstSeen = None
        self.lastSeen = None
        self.latency = None
        self.fails = None
        self.sortKey = self._calcSortKey()

    def _calcSortKey( self ) :
        notAlive = (self.fails is None) or (self.fails != 0)
        negPrio = self.isSeedNode
        if self.latency is None :
            latency = 1000
        else :
            latency = self.latency
        return (notAlive,negPrio,latency,self.sortId)

    def nodeSeen( self, latency ) :
        if self.firstSeen is None :
            self.firstSeen = self.lastSeen = time()
            self.latency = latency
            self.fails = 0
        else :
            self.lastSeen = time()
            self.latency = (self.latency + latency)/2.0
            self.fails = 0
        self.sortKey = self._calcSortKey()

    def nodeFailed( self ) :
        if self.firstSeen is not None :
            self.fails += 1
            self.sortKey = self._calcSortKey()

    def __cmp__( self, other ) :
        return cmp(self.sortKey,other.sortKey)

class NodeTable( object ) :
    def __init__( self, seedNodes ) :
        self.seedNodes = set( seedNodes )
        self.nodeList = []
        self.nodeDict = {}
        self.nextSortId = 0
        for nodeAddr in seedNodes :
            self.addNode( nodeAddr )

    def _getInfo( self, nodeAddr ) :
        info = self.nodeDict.pop( nodeAddr, None )
        if info is None :
            isSeedNode = nodeAddr in self.seedNodes
            sortId = self.nextSortId
            self.nextSortId += 1
            return _NodeInfo(nodeAddr,isSeedNode,sortId)
        pos = bisect_left( self.nodeList, info )
        del self.nodeList[pos]
        return info

    def _setInfo( self, nodeInfo ) :
        self.nodeDict[nodeInfo.nodeAddr] = nodeInfo
        insort_right( self.nodeList, nodeInfo )
        if len(self.nodeList) <= MAX_NODES : return
        extra = self.nodeList[MAX_NODES:]
        del self.nodeList[MAX_NODES:]
        for info in extra :
            del self.nodeDict[info.nodeAddr]

    def addNode( self, nodeAddr ) :
        info = self._getInfo( nodeAddr )
        self._setInfo( info )

    def nodeSeen( self, nodeAddr, latency ) :
        info = self._getInfo( nodeAddr )
        info.nodeSeen( latency )
        self._setInfo( info )

    def nodeFailed( self, nodeAddr ) :
        info = self._getInfo( nodeAddr )
        info.nodeFailed()
        self._setInfo( info )

    def getNodes( self, count=5 ) :
        return [info.nodeAddr for info in self.nodeList[:count]]

    def getLiveNodes( self, count=5 ) :
        count = min( count, len(self.nodeList) )
        out = []
        for pos in range(count) :
            info = self.nodeList[pos]
            if (info.fails is None) or (info.fails != 0) :
                break
            out.append( info.nodeAddr )
        if not out :
            out.extend( self.seedNodes )
        return out

    def getSeedNodes( self ) :
        return list(self.seedNodes)

    def getRandNode( self ) :
        return choice(self.nodeList).nodeAddr

    def getAllNodes( self ) :
        return self.getNodes( MAX_NODES )

    def getAllNonSeedNodes( self ) :
        return [info.nodeAddr for info in self.nodeList\
                if not info.isSeedNode]

class Dummy : pass

def _findLiveNodes( count, nodeTable, dhtClient, reactor, callback=None ) :
    obj = Dummy()
    def doCancel() :
        for pingOp in obj.pingOps :
            pingOp.cancel()
        obj.pingOps.clear()
        obj.timerOp.cancel()
        obj.timerOp = None
    def doNotify() :
        doCancel()
        op.notify()
    def onPing( pingOp, err, payload ) :
        obj.pingOps.remove( pingOp )
        if err >= 0 :
            obj.success += 1
        if obj.success >= count :
            doNotify()
        elif (not obj.nodes) and (not obj.pingOps) :
            doNotify()
    def doPing( nodeAddr ) :
        pingOp = dhtClient.callPing( nodeAddr,
                lambda e,p : onPing(pingOp,e,p) )
        obj.pingOps.add( pingOp )
    def doPingNodes( count ) :
        current = obj.nodes[:count]
        del obj.nodes[:count]
        for nodeAddr in current :
            doPing( nodeAddr )
    def onTimer() :
        obj.timerCount += 1
        if not obj.pingedSeed :
            if (not obj.nodes) or (obj.timerCount == \
                    FIND_LIVE_NODES_TIMEOUT_COUNT ) :
                for nodeAddr in nodeTable.getSeedNodes() :
                    doPing( nodeAddr )
                obj.pingedSeed = True
                return
        if not obj.nodes :
            return
        doPingNodes( 10 )
    obj.success = 0
    obj.pingOps = set()
    obj.nodes = nodeTable.getAllNodes()
    doPingNodes( 10 )
    obj.timerCount = 0
    obj.pingedSeed = False
    obj.timerOp = reactor.addTimer( FIND_LIVE_NODES_TIMEOUT, onTimer )
    op = AsyncOp( callback, doCancel )
    return op

def _findNewNodes( nodeTable, dhtClient, reactor, callback=None ) :
    destId = numToId( randint(0,DHT_ID_MAX-1) )
    startNodes = nodeTable.getLiveNodes()
    def onLookup( nodes ) :
        op.notify()
    lookupOp = dhtClient.lookup( destId, startNodes, onLookup )
    op = AsyncOp( callback, lookupOp.cancel )
    return op

def initNodeTable( nodeTable, dhtClient, reactor, callback=None ) :
    def onFindLive() :
        global logger
        logger.debug('done finding live nodes')
        findNewOp = _findNewNodes( nodeTable, dhtClient, reactor,
                op.notify )
        op.setCanceler( findNewOp.cancel )
    global logger
    logger.debug('finding live nodes...')
    findLiveOp = _findLiveNodes( 5, nodeTable, dhtClient, reactor,
            onFindLive )
    op = AsyncOp( callback, findLiveOp.cancel )
    return op

class NodeTableRefresher( object ) :
    def __init__( self, nodeTable, dhtClient, reactor ) :
        self.nodeTable = nodeTable
        self.dhtClient = dhtClient
        self.reactor = reactor
        self.pingOp = reactor.callLater( NODETABLE_PING_INTERVAL,
                self._onPingTimer )
        self.lookupOp = reactor.callLater( NODETABLE_LOOKUP_INTERVAL,
                self._onLookupTimer )

    def close( self ) :
        self.pingOp.cancel()
        self.lookupOp.cancel()

    def _onPing( self, err, payload ) :
        if err < 0 :
            self._onPingTimer()
            return
        self.pingOp = self.reactor.callLater( NODETABLE_PING_INTERVAL,
                self._onPingTimer )

    def _onPingTimer( self ) :
        nodeAddr = self.nodeTable.getRandNode()
        self.pingOp = self.dhtClient.callPing( nodeAddr, self._onPing )

    def _onLookup( self, nodes ) :
        self.lookupOp = self.reactor.callLater( NODETABLE_LOOKUP_INTERVAL,
                self._onLookupTimer )

    def _onLookupTimer( self ) :
        destId = numToId( randint(0,DHT_ID_MAX-1) )
        startNodes = self.nodeTable.getLiveNodes()
        self.lookupOp = self.dhtClient.lookup( destId, startNodes,
                self._onLookup )



#!/usr/bin/python
#backup_tar.py
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#
#  We want a pipe output or input so we don't need to store intermediate data.
#
#  The popen starts another process.  That process can block but we don't.
#  backup.py only takes data from this pipe when it is ready.

import os
import sys
import subprocess

try:
    import lib.dhnio as dhnio
except:
    dirpath = os.path.dirname(os.path.abspath(sys.argv[0]))
    sys.path.insert(0, os.path.abspath('datahaven'))
    sys.path.insert(0, os.path.abspath(os.path.join(dirpath, '..')))
    sys.path.insert(0, os.path.abspath(os.path.join(dirpath, '..', '..')))
    try:
        import lib.dhnio as dhnio
    except:
        sys.exit()

import lib.nonblocking as nonblocking

#------------------------------------------------------------------------------ 

def run(cmdargs):
    dhnio.Dprint(14, "backup_tar.run %s" % str(cmdargs))
    try:
        if dhnio.Windows():
            import win32process
            p = nonblocking.Popen(
                cmdargs,
                shell=False,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=False,
                creationflags = win32process.CREATE_NO_WINDOW,)
        else:
            p = nonblocking.Popen(
                cmdargs,
                shell=False,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=False,)
    except:
        dhnio.Dprint(1, 'backup_tar.run ERROR executing: ' + str(cmdargs) + '\n' + str(dhnio.formatExceptionInfo()))
        return None
    return p


# Returns file descriptor for process that makes tar archive
def backuptar(directorypath, recursive_subfolders=True, compress=None):
    if not os.path.isdir(directorypath):
        dhnio.Dprint(1, 'backup_tar.backuptar ERROR %s not found' % directorypath)
        return None
    subdirs = 'subdirs'
    if not recursive_subfolders:
        subdirs = 'nosubdirs'
    if compress is None:
        compress = 'none'
    # dhnio.Dprint(14, "backup_tar.backuptar %s %s compress=%s" % (directorypath, subdirs, compress))
    if dhnio.Windows():
        if dhnio.isFrozen():
            commandpath = "dhnbackup.exe"
            cmdargs = [commandpath, subdirs, compress, directorypath]
        else:
            commandpath = "dhnbackup.py"
            cmdargs = [sys.executable, commandpath, subdirs, compress, directorypath]
    else:
        commandpath = "dhnbackup.py"
        cmdargs = [sys.executable, commandpath, subdirs, compress, directorypath]
    if not os.path.isfile(commandpath):
        dhnio.Dprint(1, 'backup_tar.backuptar ERROR %s not found' % commandpath)
        return None
    # dhnio.Dprint(14, "backup_tar.backuptar going to execute %s" % str(cmdargs))
    p = run(cmdargs)
    return p


# Returns file descriptor for process that makes tar archive
def backuptarfile(filepath, compress=None):
    if not os.path.isfile(filepath):
        dhnio.Dprint(1, 'backup_tar.backuptarfile ERROR %s not found' % filepath)
        return None
    if compress is None:
        compress = 'none'
    # dhnio.Dprint(14, "backup_tar.backuptarfile %s compress=%s" % (filepath, compress))
    if dhnio.Windows():
        if dhnio.isFrozen():
            commandpath = "dhnbackup.exe"
            cmdargs = [commandpath, 'nosubdirs', compress, filepath]
        else:
            commandpath = "dhnbackup.py"
            cmdargs = [sys.executable, commandpath, 'nosubdirs', compress, filepath]
    else:
        commandpath = "dhnbackup.py"
        cmdargs = [sys.executable, commandpath, 'nosubdirs', compress, filepath]
    if not os.path.isfile(commandpath):
        dhnio.Dprint(1, 'backup_tar.backuptarfile ERROR %s not found' % commandpath)
        return None
    # dhnio.Dprint(12, "backup_tar.backuptarfile going to execute %s" % str(cmdargs))
    p = run(cmdargs)
    return p


def extracttar(tarfile, outdir):
    if not os.path.isfile(tarfile):
        dhnio.Dprint(1, 'backup_tar.extracttar ERROR %s not found' % tarfile)
        return None
    # dhnio.Dprint(12, "backup_tar.extracttar %s %s" % (tarfile, outdir))
    if dhnio.Windows():
        if dhnio.isFrozen():
            commandpath = 'dhnbackup.exe'
            cmdargs = [commandpath, 'extract', tarfile, outdir]
        else:
            commandpath = "dhnbackup.py"
            cmdargs = [sys.executable, commandpath, 'extract', tarfile, outdir]
    else:
        commandpath = "dhnbackup.py"
        cmdargs = [sys.executable, commandpath, 'extract', tarfile, outdir]
    if not os.path.isfile(commandpath):
        dhnio.Dprint(1, 'backup_tar.extracttar ERROR %s is not found' % commandpath)
        return None
    p = run(cmdargs)
    return p


#------------------------------------------------------------------------------ 

if __name__ == "__main__":
    dhnio.SetDebug(20)
    p = backuptar(sys.argv[1])
    p.make_nonblocking()
    print p
    print p.wait()
    



#!/usr/bin/python
#dhntester.py

import os
import sys
import platform

import time

#------------------------------------------------------------------------------ 

def logfilepath():
    if platform.uname()[0] == 'Windows':
        logspath = os.path.join(os.environ['APPDATA'], 'DataHaven.NET', 'logs')
    else:
        logspath = os.path.join(os.path.expanduser('~'), '.datahaven', 'logs')
    if not os.path.isdir(logspath):
        return 'dhntester.log'
    return os.path.join(logspath, 'dhntester.log')

def printlog(txt):
    LogFile = open(logfilepath(), 'a')
    LogFile.write(txt+'\n')
    LogFile.close()
    
#------------------------------------------------------------------------------ 

if __name__ == "__main__":
    dirpath = os.path.dirname(os.path.abspath(sys.argv[0]))
    sys.path.insert(0, os.path.abspath('datahaven'))
    sys.path.insert(0, os.path.abspath(os.path.join(dirpath, '..')))
    sys.path.insert(0, os.path.abspath(os.path.join(dirpath, '..', '..')))

try:
    import lib.dhnio as dhnio
    from lib.dhnpacket import Unserialize
    from lib.nameurl import FilenameUrl
    from lib.settings import init as settings_init
    from lib.settings import CustomersSpaceFile, getCustomersFilesDir, LocalTesterLogFilename
    from lib.settings import BackupIndexFileName
    from lib.contacts import init as contacts_init
    from lib.commands import init as commands_init
except:
    import traceback
    printlog(traceback.format_exc())
    sys.exit(2)

# need to make sure the dhntester log is in a directory the user has permissions for,
# such as the customer data directory.  Possibly move to temp directory?
# sys.stdout = myoutput
# sys.stderr = myoutput

#-------------------------------------------------------------------------------
#test all packets for each customer.
#check if he use more space than we gave him and if packets is too old.
def SpaceTime():
    printlog('SpaceTime ' + str(time.strftime("%a, %d %b %Y %H:%M:%S +0000")))
    space = dhnio._read_dict(CustomersSpaceFile())
    if space is None:
        printlog('SpaceTime ERROR can not read file ' + CustomersSpaceFile())
        return
    customers_dir = getCustomersFilesDir()
    if not os.path.exists(customers_dir):
        printlog('SpaceTime ERROR customers folder not exist')
        return
    remove_list = {}
    for customer_filename in os.listdir(customers_dir):
        onecustdir = os.path.join(customers_dir, customer_filename)
        if not os.path.isdir(onecustdir):
            remove_list[onecustdir] = 'is not a folder'
            continue
        idurl = FilenameUrl(customer_filename)
        if idurl is None:
            remove_list[onecustdir] = 'wrong folder name'
            continue
        curspace = space.get(idurl, None)
        if curspace is None:
            continue
        try:
            maxspaceV = int(float(curspace) * 1024 * 1024) #in bytes
        except:
            remove_list[onecustdir] = 'wrong space value'
            continue
        timedict = {}
        sizedict = {}
        def cb(path, subpath, name):
            if not os.access(path, os.R_OK | os.W_OK):
                return False
            if not os.path.isfile(path):
                return True
            if name in [BackupIndexFileName(),]:
                return False
            stats = os.stat(path)
            timedict[path] = stats.st_ctime
            sizedict[path] = stats.st_size
        dhnio.traverse_dir_recursive(cb, onecustdir)
        currentV = 0
        for path in sorted(timedict.keys(), key=lambda x:timedict[x], reverse=True):
            currentV += sizedict.get(path, 0)
            if currentV < maxspaceV:
                continue
            try:
                os.remove(path)
                printlog('SpaceTime ' + path + ' file removed (cur:%s, max: %s)' % (str(currentV), str(maxspaceV)) )
            except:
                printlog('SpaceTime ERROR removing ' + path)
            time.sleep(0.1)
        timedict.clear()
        sizedict.clear()
    for path in remove_list.keys():
        if not os.path.exists(path):
            continue
        if os.path.isdir(path):
            try:
                dhnio._dir_remove(path)
                printlog('SpaceTime ' + path + ' dir removed (%s)' % (remove_list[path]))
            except:
                printlog('SpaceTime ERROR removing ' + path)
            continue
        if not os.access(path, os.W_OK):
            os.chmod(path, 0600)
        try:
            os.remove(path)
            printlog('SpaceTime ' + path + ' file removed (%s)' % (remove_list[path]))
        except:
            printlog('SpaceTime ERROR removing ' + path)
    del remove_list

#------------------------------------------------------------------------------
#test packets after list of customers was changed
def UpdateCustomers():
    printlog('UpdateCustomers ' + str(time.strftime("%a, %d %b %Y %H:%M:%S +0000")))
    space = dhnio._read_dict(CustomersSpaceFile())
    if space is None:
        printlog('UpdateCustomers ERROR space file can not read' )
        return
    customers_dir = getCustomersFilesDir()
    if not os.path.exists(customers_dir):
        printlog('UpdateCustomers ERROR customers folder not exist')
        return
    remove_list = {}
    for customer_filename in os.listdir(customers_dir):
        onecustdir = os.path.join(customers_dir, customer_filename)
        if not os.path.isdir(onecustdir):
            remove_list[onecustdir] = 'is not a folder'
            continue
        idurl = FilenameUrl(customer_filename)
        if idurl is None:
            remove_list[onecustdir] = 'wrong folder name'
            continue
        curspace = space.get(idurl, None)
        if curspace is None:
            remove_list[onecustdir] = 'is not a customer'
            continue
    for path in remove_list.keys():
        if not os.path.exists(path):
            continue
        if os.path.isdir(path):
            try:
                dhnio._dir_remove(path)
                printlog('UpdateCustomers ' + path + ' folder removed (%s)' % (remove_list[path]))
            except:
                printlog('UpdateCustomers ERROR removing ' + path)
            continue
        if not os.access(path, os.W_OK):
            os.chmod(path, 0600)
        try:
            os.remove(path)
            printlog('UpdateCustomers ' + path + ' file removed (%s)' % (remove_list[path]))
        except:
            printlog('UpdateCustomers ERROR removing ' + path)

#------------------------------------------------------------------------------
#check all packets to be valid
def Validate():
    printlog('Validate ' + str(time.strftime("%a, %d %b %Y %H:%M:%S +0000")))
    contacts_init()
    commands_init()
    customers_dir = getCustomersFilesDir()
    if not os.path.exists(customers_dir):
        return
    for customer_filename in os.listdir(customers_dir):
        onecustdir = os.path.join(customers_dir, customer_filename)
        if not os.path.isdir(onecustdir):
            continue
        def cb(path, subpath, name):
            if not os.access(path, os.R_OK | os.W_OK):
                return False
            if not os.path.isfile(path):
                return True
            if name in [BackupIndexFileName(),]:
                return False
            packetsrc = dhnio.ReadBinaryFile(path)
            if not packetsrc:
                try:
                    os.remove(path) # if is is no good it is of no use to anyone
                    printlog('Validate ' + path + ' removed (empty file)')
                except:
                    printlog('Validate ERROR removing ' + path)
                    return False
            packet = Unserialize(packetsrc)
            if packet is None:
                try:
                    os.remove(path) # if is is no good it is of no use to anyone
                    printlog('Validate ' + path + ' removed (unserialize error)')
                except:
                    printlog('Validate ERROR removing ' + path)
                    return False
            result = packet.Valid()
            packetsrc = ''
            del packet
            if not result:
                try:
                    os.remove(path) # if is is no good it is of no use to anyone
                    printlog('Validate ' + path + ' removed (invalid packet)')
                except:
                    printlog('Validate ERROR removing ' + path)
                    return False
            time.sleep(0.1)
            return False
        dhnio.traverse_dir_recursive(cb, onecustdir)

#------------------------------------------------------------------------------

def main():
    if len(sys.argv) < 2:
        return
    dhnio.init()
    dhnio.DisableLogs()
    dhnio.DisableOutput()
    settings_init()
    dhnio.SetDebug(0)
    dhnio.LowerPriority()
#    dhnio.OpenLogFile(LocalTesterLogFilename(), True)
#    dhnio.StdOutRedirectingStart()
#    dhnio.LifeBegins()
    commands = {
        'update_customers' : UpdateCustomers,
        'validate' : Validate,
        'space_time' : SpaceTime,
    }
    cmd = commands.get(sys.argv[1], None)
    if not cmd:
        printlog('ERROR wrong command: ' + str(sys.argv))
        return
    cmd()
#    dhnio.StdOutRedirectingStop()
#    dhnio.CloseLogFile()

#------------------------------------------------------------------------------ 

if __name__ == "__main__":
    main()









from socket import inet_aton
from struct import pack, unpack
from types import IntType, StringType, TupleType
from ncrypt0.digest import DigestType, Digest
from ncrypt0.rsa import RSAKey, RSAError
from nitro.bencode import encode
from cspace.dht.params import DHT_ID_LENGTH, DHT_ID_MAX
from cspace.util.hexcode import hexDecode, hexEncode

digestType = DigestType( 'SHA1' )
digestLength = digestType.size()
assert digestLength == DHT_ID_LENGTH

def toId( x ) :
    return Digest(digestType).digest(x)

def idToNum( x ) :
    return long(x.encode('hex'),16)

def numToId( numId ) :
    s = hex( numId )
    assert s.startswith('0x')
    if s.endswith('L') : s = s[2:-1]
    else : s = s[2:]
    if len(s) < 2*DHT_ID_LENGTH :
        s = ('0'*(2*DHT_ID_LENGTH-len(s))) + s
    x = s.decode('hex')
    assert len(x) == DHT_ID_LENGTH
    return x

def checkIP( ip ) :
    if type(ip) is not StringType : return False
    if not ip : return False
    try :
        inet_aton( ip )
        return True
    except :
        return False

def checkPort( port ) :
    if type(port) is not IntType : return False
    return 0 < port < 65536

def checkAddr( addr ) :
    if type(addr) is not TupleType :
        return False
    if len(addr) != 2 : return False
    if not checkIP(addr[0]) : return False
    return checkPort( addr[1] )

def addrToStr( addr ) :
    ip = unpack( 'I', inet_aton(addr[0]) )[0]
    port = addr[1]
    return pack( '!IH', ip, port )

def addrToId( addr ) :
    return toId( addrToStr(addr) )

def verifySignature( publicKey, data, updateLevel, signature ) :
    payload = encode( ('DHT-DATA',data,updateLevel) )
    if type(publicKey) is str :
        k = RSAKey()
        try :
            k.fromDER_PublicKey( publicKey )
        except RSAError :
            return False
    else :
        k = publicKey
    try :
        digest = Digest(digestType).digest( payload )
        k.verify( signature, digest, digestType )
        return True
    except RSAError :
        return False

def computeSignature( rsaKey, data, updateLevel ) :
    payload = encode( ('DHT-DATA',data,updateLevel) )
    digest = Digest(digestType).digest( payload )
    sig = rsaKey.sign( digest, digestType )
    return str(sig[0])






#!/usr/bin/python
#packetid.py
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#
# packetid  - we have a standard way of making the PacketID strings for many packets.

# Note we return strings and not integers

import time
import re

#------------------------------------------------------------------------------ 

_LastUniqueNumber = 0

#------------------------------------------------------------------------------ 

def UniqueID():
    global _LastUniqueNumber
    _LastUniqueNumber += 1
    # we wrap around every billion, old packets should be gone by then
    if _LastUniqueNumber > 1000000000:     
        _LastUniqueNumber = 0
    inttime = int(time.time() * 100.0)
    if _LastUniqueNumber < inttime:
        _LastUniqueNumber = inttime
    # strings for packet fields
    return str(_LastUniqueNumber) 

def MakePacketID(backupID, blockNumber, supplierNumber, dataORparity):
    return backupID + '/' + str(blockNumber) + '-' + str(supplierNumber) + '-' + dataORparity

def Valid(packetID):
    # packetID may be in different forms:
    #    + full:     0/0/1/0/F20131120053803PM/0-1-Data
    #    + backupID: 0/0/1/0/F20131120053803PM
    #    + pathID:   0/0/1/0
    # here is:
    #     pathID:        0/0/1/0
    #     versionName:   F20131120053803PM
    #     blockNum :     0
    #     supplierNum :  1
    #     dataORparity : Data
    head, x, tail = packetID.rpartition('/')
    if x == '' and head == '':
        # this seems to be a shortest pathID: 0, 1, 2, ...
        try:
            x = int(tail)
        except:
            return False
        return True
    if tail.endswith('-Data') or tail.endswith('-Parity'):
        # seems to be in the full form
        if not IsPacketNameCorrect(tail):
            return False
        pathID, x, versionName = head.rpartition('/')
        if not IsCanonicalVersion(versionName):
            return False
        if not IsPathIDCorrect(pathID):
            return False
        return True
    if IsPathIDCorrect(packetID):
        # we have only two options now, let's if this is a pathID
        return True
        # this should be a backupID - no other cases 
    if IsCanonicalVersion(tail) and IsPathIDCorrect(head):
        return True
    # something is not fine with that string
    return False

def Split(packetID):
    # packetID :     0/0/1/0/F20131120053803PM/0-1-Data
    # return:        "0/0/1/0/F20131120053803PM", "0", "1", "Data"
    try:
        backupID, x, fileName = packetID.rpartition('/')
        blockNum, supplierNum, dataORparity = fileName.split('-')
        blockNum = int(blockNum)
        supplierNum = int(supplierNum)
    except:
        return None, None, None, None
    return backupID, blockNum, supplierNum, dataORparity

def SplitFull(packetID):
    # packetID :     0/0/1/0/F20131120053803PM/0-1-Data
    # return:        "0/0/1/0", "F20131120053803PM", "0", "1", "Data"
    try:
        backupID, x, fileName = packetID.rpartition('/')
        pathID, x, versionName = backupID.rpartition('/')
        blockNum, supplierNum, dataORparity = fileName.split('-')
        blockNum = int(blockNum)
        supplierNum = int(supplierNum)
    except:
        return None, None, None, None, None
    return pathID, versionName, blockNum, supplierNum, dataORparity

def SplitVersionFilename(packetID): 
    # packetID :     0/0/1/0/F20131120053803PM/0-1-Data
    # return:        "0/0/1/0", "F20131120053803PM", "0-1-Data"
    try:
        backupID, x, fileName = packetID.rpartition('/')
        pathID, x, versionName = backupID.rpartition('/')
    except:
        return None, None, None
    return pathID, versionName, fileName

def SplitBackupID(backupID):
    # backupID :     0/0/1/0/F20131120053803PM
    # return :       "0/0/1/0", "F20131120053803PM"
    try:
        pathID, x, versionName = backupID.rpartition('/')
    except:
        return None, None
    return pathID, versionName

def IsCanonicalVersion(versionName):
    return re.match('^F\d+?(AM|PM)\d*?$', versionName) is not None

def IsPacketNameCorrect(fileName):
    return re.match('^\d+?\-\d+?\-(Data|Parity)$', fileName) is not None

def IsPathIDCorrect(pathID):
    return pathID.replace('/', '').isdigit()

def BidBnSnDp(packetID):
    return Split(packetID)

def BackupID(packetID):
    return Split(packetID)[0]

def BlockNumber(packetID):
    return Split(packetID)[1]

def SupplierNumber(packetID):
    return Split(packetID)[2]

def DataOrParity(packetID):
    return Split(packetID)[3]

def parentPathsList(ID):
    # return all parent paths of the given ID
    # ID: 0/0/1/0/F20131120053803PM/0-1-Data
    # will return: 
    # [ '0', '0/0', '0/0/1', '0/0/1/0', '0/0/1/0/F20131120053803PM', 0/0/1/0/F20131120053803PM/0-1-Data ]
    path = '' 
    for word in ID.split('/'):
        if path:
            path += '/'
        path += word
        yield path
        




